import logging
from collections.abc import Sequence
from pathlib import Path

from pipx.colors import bold
from pipx.constants import ExitCode
from pipx.emojis import sleep
from pipx.util import PipxError
from pipx.venv import Venv

logger = logging.getLogger(__name__)


def _update_pin_info(venv: Venv, package_name: str, is_main_package: bool, unpin: bool) -> int:
    package_metadata = venv.package_metadata[package_name]
    venv.update_package_metadata(
        package_name=str(package_metadata.package),
        package_or_url=str(package_metadata.package_or_url),
        pip_args=package_metadata.pip_args,
        include_dependencies=package_metadata.include_dependencies,
        include_apps=package_metadata.include_apps,
        is_main_package=is_main_package,
        suffix=package_metadata.suffix,
        pinned=not unpin,
    )
    return 1


def pin(
    venv_dir: Path,
    verbose: bool,
    skip: Sequence[str],
    injected_only: bool = False,
) -> ExitCode:
    venv = Venv(venv_dir, verbose=verbose)
    try:
        main_package_metadata = venv.package_metadata[venv.main_package_name]
    except KeyError as e:
        raise PipxError(f"Package {venv.name} is not installed") from e

    if injected_only or skip:
        pinned_packages_count = 0
        pinned_packages_list = []
        for package_name in venv.package_metadata:
            if package_name == venv.main_package_name or package_name in skip:
                continue

            if venv.package_metadata[package_name].pinned:
                print(f"{package_name} was pinned. Not modifying.")
                continue

            pinned_packages_count += _update_pin_info(venv, package_name, is_main_package=False, unpin=False)
            pinned_packages_list.append(f"{package_name} {venv.package_metadata[package_name].package_version}")

        if pinned_packages_count != 0:
            print(bold(f"Pinned {pinned_packages_count} packages in venv {venv.name}"))
            for package in pinned_packages_list:
                print("  -", package)
    elif main_package_metadata.pinned:
        logger.warning(f"Package {main_package_metadata.package} already pinned {sleep}")
    else:
        for package_name in venv.package_metadata:
            if package_name == venv.main_package_name:
                _update_pin_info(venv, venv.main_package_name, is_main_package=True, unpin=False)
            else:
                _update_pin_info(venv, package_name, is_main_package=False, unpin=False)

    return ExitCode(0)


def unpin(venv_dir: Path, verbose: bool) -> ExitCode:
    venv = Venv(venv_dir, verbose=verbose)
    try:
        main_package_metadata = venv.package_metadata[venv.main_package_name]
    except KeyError as e:
        raise PipxError(f"Package {venv.name} is not installed") from e

    unpinned_packages_count = 0
    unpinned_packages_list = []

    for package_name in venv.package_metadata:
        if package_name == main_package_metadata.package and main_package_metadata.pinned:
            unpinned_packages_count += _update_pin_info(venv, package_name, is_main_package=True, unpin=True)
        elif venv.package_metadata[package_name].pinned:
            unpinned_packages_count += _update_pin_info(venv, package_name, is_main_package=False, unpin=True)
        unpinned_packages_list.append(package_name)

    if unpinned_packages_count != 0:
        print(bold(f"Unpinned {unpinned_packages_count} packages in venv {venv.name}"))
        for package in unpinned_packages_list:
            print("  -", package)
    else:
        logger.warning(f"No packages to unpin in venv {venv.name}")

    return ExitCode(0)



import enum
import os
import platform
import sysconfig
from textwrap import dedent
from typing import NewType


# enum.StrEnum is Python 3.11+; pipx supports 3.10.
class FetchPythonOptions(str, enum.Enum):
    ALWAYS = "always"
    MISSING = "missing"
    NEVER = "never"

    def __str__(self) -> str:
        return self.value


PIPX_SHARED_PTH = "pipx_shared.pth"
TEMP_VENV_EXPIRATION_THRESHOLD_DAYS = 14
MINIMUM_PYTHON_VERSION = "3.10"
MAN_SECTIONS = [f"man{i}" for i in range(1, 10)]


_FETCH_MISSING_PYTHON_RAW = os.environ.get("PIPX_FETCH_MISSING_PYTHON")
_FETCH_PYTHON_RAW = os.environ.get("PIPX_FETCH_PYTHON")
_FALSY = ("", "0", "false", "f", "no", "n", "off")


def _compute_fetch_python(missing_raw: str | None, python_raw: str | None) -> tuple[FetchPythonOptions, bool]:
    fetch_missing = missing_raw is not None and missing_raw.strip().lower() not in _FALSY
    raw = python_raw or (FetchPythonOptions.MISSING.value if fetch_missing else FetchPythonOptions.NEVER.value)
    try:
        return FetchPythonOptions(raw), False
    except ValueError:
        return FetchPythonOptions.NEVER, True


_FETCH_PYTHON, _FETCH_PYTHON_INVALID = _compute_fetch_python(_FETCH_MISSING_PYTHON_RAW, _FETCH_PYTHON_RAW)


ExitCode = NewType("ExitCode", int)
# pipx shell exit codes
EXIT_CODE_OK = ExitCode(0)
EXIT_CODE_INJECT_ERROR = ExitCode(1)
EXIT_CODE_UNINJECT_ERROR = ExitCode(1)
EXIT_CODE_INSTALL_VENV_EXISTS = ExitCode(0)
EXIT_CODE_LIST_PROBLEM = ExitCode(1)
EXIT_CODE_UNINSTALL_VENV_NONEXISTENT = ExitCode(1)
EXIT_CODE_UNINSTALL_ERROR = ExitCode(1)
EXIT_CODE_REINSTALL_VENV_NONEXISTENT = ExitCode(1)
EXIT_CODE_REINSTALL_INVALID_PYTHON = ExitCode(1)
EXIT_CODE_SPECIFIED_PYTHON_EXECUTABLE_NOT_FOUND = ExitCode(1)


def is_windows() -> bool:
    return platform.system() == "Windows"


def is_macos() -> bool:
    return platform.system() == "Darwin"


def is_linux() -> bool:
    return platform.system() == "Linux"


def is_mingw() -> bool:
    return sysconfig.get_platform().startswith("mingw")


WINDOWS: bool = is_windows()
MACOS: bool = is_macos()
LINUX: bool = is_linux()
MINGW: bool = is_mingw()

completion_instructions = dedent(
    """
If you encountered register-python-argcomplete command not found error,
or if you are using zipapp, run

    pipx install argcomplete

before running any of the following commands.

Add the appropriate command to your shell's config file
so that it is run on startup. You will likely have to restart
or re-login for the autocompletion to start working.

bash:
    eval "$(register-python-argcomplete pipx)"

zsh:
    To activate completions in zsh, first make sure compinit is marked for
    autoload and run autoload:

    autoload -U compinit && compinit

    Afterwards you can enable completions for pipx:

    eval "$(register-python-argcomplete pipx)"

    NOTE: If your version of argcomplete is earlier than v3, you may need to
    have bashcompinit enabled in zsh by running:

    autoload -U bashcompinit
    bashcompinit


tcsh:
    eval `register-python-argcomplete --shell tcsh pipx`

fish:
    # Not required to be in the config file, only run once
    register-python-argcomplete --shell fish pipx >~/.config/fish/completions/pipx.fish

"""
)



import sys
from collections.abc import Callable

try:
    import colorama  # type: ignore[import-untyped]
except ImportError:  # Colorama is Windows only package
    colorama = None

PRINT_COLOR = bool(sys.stdout and sys.stdout.isatty())

if PRINT_COLOR and colorama:
    colorama.init()


class c:
    header = "\033[95m"
    blue = "\033[94m"
    green = "\033[92m"
    yellow = "\033[93m"
    red = "\033[91m"
    bold = "\033[1m"
    cyan = "\033[96m"
    underline = "\033[4m"
    end = "\033[0m"


def mkcolorfunc(style: str) -> Callable[[str], str]:
    def stylize_text(x: str) -> str:
        if PRINT_COLOR:
            return f"{style}{x}{c.end}"
        else:
            return x

    return stylize_text


bold = mkcolorfunc(c.bold)
red = mkcolorfunc(c.red)
blue = mkcolorfunc(c.cyan)
cyan = mkcolorfunc(c.blue)
green = mkcolorfunc(c.green)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Final

if TYPE_CHECKING:
    from pathlib import Path
    from subprocess import CompletedProcess


PIP: Final[str] = "pip"
UV: Final[str] = "uv"
KNOWN_BACKENDS: Final[tuple[str, ...]] = (PIP, UV)


class Backend(ABC):
    name: str

    @abstractmethod
    def create_venv(
        self,
        root: Path,
        *,
        python: str,
        venv_args: list[str],
        pip_args: list[str],
        include_pip: bool,
        verbose: bool,
    ) -> None: ...

    @abstractmethod
    def install(
        self,
        *,
        venv_root: Path,
        venv_python: Path,
        requirements: list[str],
        pip_args: list[str],
        no_deps: bool = False,
        upgrade: bool = False,
        log_pip_errors: bool = True,
        verbose: bool = False,
    ) -> CompletedProcess[str]: ...

    @abstractmethod
    def uninstall(
        self,
        *,
        venv_root: Path,
        venv_python: Path,
        package: str,
        verbose: bool,
    ) -> CompletedProcess[str]: ...

    @abstractmethod
    def list_installed(
        self,
        *,
        venv_root: Path,
        venv_python: Path,
        not_required: bool = False,
    ) -> set[str]: ...

    @abstractmethod
    def run_raw_pip(
        self,
        *,
        venv_root: Path,
        venv_python: Path,
        args: list[str],
        capture_stdout: bool = True,
        capture_stderr: bool = True,
        verbose: bool = False,
    ) -> CompletedProcess[str]: ...

    @abstractmethod
    def needs_shared_libs(self) -> bool: ...

    @abstractmethod
    def upgrade_packaging_libraries(
        self,
        venv_python: Path,
        pip_args: list[str],
        *,
        verbose: bool,
    ) -> None: ...


__all__ = [
    "KNOWN_BACKENDS",
    "PIP",
    "UV",
    "Backend",
]



from __future__ import annotations

import os
from functools import cache

from pipx.backends._base import KNOWN_BACKENDS, PIP, UV, Backend
from pipx.backends.pip import PipBackend
from pipx.backends.uv import UvBackend, find_uv_binary
from pipx.util import PipxError


def resolve_backend_name(
    *,
    cli_value: str | None = None,
    env_value: str | None = None,
    metadata_value: str | None = None,
    auto: bool = True,
) -> tuple[str, str]:
    """Return ``(name, source)`` per precedence cli > metadata > env > auto.

    ``metadata`` sits above ``env`` so ``PIPX_DEFAULT_BACKEND=uv`` cannot
    silently retarget pip-backed venvs already on disk.
    """
    for candidate, source in ((cli_value, "cli"), (metadata_value, "metadata"), (env_value, "env")):
        if (validated := _validate(candidate)) is not None:
            return validated, source
    if auto and (binary_source := find_uv_binary()[1]) != "missing":
        return UV, f"auto-{binary_source}"
    return PIP, "auto-pip"


@cache
def get_backend(name: str) -> Backend:
    # Cached so ``UvBackend.__init__``'s version probe and log line fire once
    # per process even when validation + construction both ask for the backend.
    if name == PIP:
        return PipBackend()
    if name == UV:
        return UvBackend()
    raise PipxError(f"Unknown backend {name!r}. Valid backends: {', '.join(KNOWN_BACKENDS)}.")


def assert_not_pip_under_uv(package_name: str, backend_name: str) -> None:
    # Named narrowly: a future backend with its own deny-list should grow its
    # own guard rather than overload this one.
    if package_name == "pip" and backend_name != PIP:
        raise PipxError(
            "The 'pip' package cannot be installed or exposed via the uv backend, since uv venvs "
            "don't ship pip and the resulting environment would be inconsistent.\n"
            "Use `--backend pip` (or set PIPX_DEFAULT_BACKEND=pip) to install 'pip' as a tool."
        )


def env_default_backend() -> str | None:
    raw = os.environ.get("PIPX_DEFAULT_BACKEND")
    return raw.strip() if raw else None


def _validate(candidate: str | None) -> str | None:
    if not candidate:
        return None
    if candidate not in KNOWN_BACKENDS:
        raise PipxError(f"Unknown backend {candidate!r}. Valid backends: {', '.join(KNOWN_BACKENDS)}.")
    return candidate


__all__ = [
    "KNOWN_BACKENDS",
    "PIP",
    "UV",
    "Backend",
    "assert_not_pip_under_uv",
    "env_default_backend",
    "find_uv_binary",
    "get_backend",
    "resolve_backend_name",
]



"""``pipx run`` delegation under the uv backend, kept out of ``run.py`` so the
flag-translation surface doesn't bury the pip flow."""

from __future__ import annotations

import logging
from shutil import which
from typing import TYPE_CHECKING, Final, NoReturn

from pipx.backends.uv import resolve_uv_binary
from pipx.emojis import hazard
from pipx.util import PipxError, exec_app, pipx_wrap

if TYPE_CHECKING:
    from pathlib import Path

_LOGGER: Final[logging.Logger] = logging.getLogger(__name__)

# Pip flags translatable to ``uv tool run`` / ``uv run``; anything else errors.
_UV_TRANSLATABLE_VALUE_FLAGS: Final[frozenset[str]] = frozenset(
    {"--index-url", "-i", "--extra-index-url", "--find-links", "-f", "--no-binary", "--only-binary", "--trusted-host"}
)
_UV_TRANSLATABLE_BOOL_FLAGS: Final[dict[str, str]] = {
    "--pre": "--prerelease=allow",
    "--no-deps": "--no-deps",
    "--no-cache-dir": "--no-cache",
    "--upgrade": "--upgrade",
    "-U": "--upgrade",
}


def _reject_venv_args(venv_args: list[str]) -> None:
    # uv builds its own venv internally; silently dropping ``--venv-args``
    # (e.g. ``--system-site-packages``) would diverge from the pip path.
    if venv_args:
        raise PipxError(
            f"--venv-args ({' '.join(venv_args)}) is not supported by `pipx run --backend uv`.\n"
            "Use `pipx run --backend pip` if those flags are required, or drop them."
        )


def run_via_uv_tool_run(
    *,
    app: str,
    package_or_url: str,
    dependencies: list[str],
    app_args: list[str],
    python: str,
    pip_args: list[str],
    venv_args: list[str],
    use_cache: bool,
    verbose: bool,
    no_path_check: bool = False,
) -> NoReturn:
    _reject_venv_args(venv_args)
    if not no_path_check and (existing_app_path := which(app)):
        _LOGGER.warning(
            pipx_wrap(
                f"""
                {hazard}  {app} is already on your PATH and installed at
                {existing_app_path}. Downloading and running anyway.
                """,
                subsequent_indent=" " * 4,
            )
        )

    cmd: list[str] = [str(resolve_uv_binary()), "tool", "run"]
    if package_or_url and package_or_url != app:
        cmd += ["--from", package_or_url]
    if python:
        cmd += ["--python", python]
    for dependency in dependencies:
        cmd += ["--with", dependency]
    if not use_cache:
        cmd.append("--no-cache")
    if verbose:
        cmd.append("--verbose")
    cmd += translate_pip_args_for_uv(pip_args)
    cmd.append(app)
    cmd += app_args
    exec_app(cmd)


def run_script_via_uv_run(
    *,
    script_path: Path,
    app_args: list[str],
    python: str,
    pip_args: list[str],
    venv_args: list[str],
    use_cache: bool,
    verbose: bool,
    dependencies: list[str] | None = None,
) -> NoReturn:
    _reject_venv_args(venv_args)
    cmd: list[str] = [str(resolve_uv_binary()), "run", "--script"]
    if python:
        cmd += ["--python", python]
    if not use_cache:
        cmd.append("--no-cache")
    if verbose:
        cmd.append("--verbose")
    for dependency in dependencies or []:
        cmd += ["--with", dependency]
    cmd += translate_pip_args_for_uv(pip_args)
    cmd += [str(script_path), *app_args]
    exec_app(cmd)


def translate_pip_args_for_uv(pip_args: list[str]) -> list[str]:
    """Translate ``pip_args`` for ``uv tool run`` / ``uv run --script``.

    Strict on this boundary because ``uv tool run`` exposes a smaller flag
    surface than ``uv pip install``; the install path stays permissive (it
    forwards everything to uv) since over-rejecting valid uv pip flags would
    be more painful than the asymmetry.
    """
    translated: list[str] = []
    iterator = iter(pip_args)
    for arg in iterator:
        if arg in ("-q", "-qq", "--quiet"):
            continue
        if arg in ("--editable", "-e"):
            raise PipxError(
                "`--editable` is not supported by `pipx run --backend uv`.\n"
                "Use `pipx run --backend pip` for editable installs."
            )
        if (translated_bool := _UV_TRANSLATABLE_BOOL_FLAGS.get(arg)) is not None:
            translated.append(translated_bool)
            continue
        if arg in _UV_TRANSLATABLE_VALUE_FLAGS:
            try:
                value = next(iterator)
            except StopIteration as exc:
                raise PipxError(f"Missing value for {arg!r} in --pip-args.") from exc
            translated.extend([arg, value])
            continue
        # Bool flags must not accept ``=value`` (caught ``--pre=foo`` slipping through).
        if "=" in arg and arg.split("=", 1)[0] in _UV_TRANSLATABLE_VALUE_FLAGS:
            translated.append(arg)
            continue
        raise PipxError(
            f"--pip-args contains {arg!r}, which has no `uv tool run` equivalent.\n"
            "Use `--backend pip` if you need pip-only flags."
        )
    return translated


__all__ = [
    "run_script_via_uv_run",
    "run_via_uv_tool_run",
    "translate_pip_args_for_uv",
]



import logging
import shutil
import time
from collections.abc import Generator
from pathlib import Path
from typing import TYPE_CHECKING, Final, NoReturn

if TYPE_CHECKING:
    from subprocess import CompletedProcess

try:
    from importlib.metadata import Distribution, EntryPoint
except ImportError:
    from importlib_metadata import Distribution, EntryPoint  # type: ignore[import-not-found,no-redef]

from packaging.utils import canonicalize_name

from pipx.animate import animate
from pipx.backends import Backend, assert_not_pip_under_uv, env_default_backend, get_backend, resolve_backend_name
from pipx.constants import PIPX_SHARED_PTH, ExitCode
from pipx.emojis import hazard
from pipx.interpreter import DEFAULT_PYTHON
from pipx.package_specifier import (
    fix_package_name,
    get_extras,
    parse_specifier_for_install,
    parse_specifier_for_metadata,
)
from pipx.pipx_metadata_file import PackageInfo, PipxMetadata
from pipx.shared_libs import (
    DISABLE_SHARED_LIBS_AUTO_UPGRADE,
    shared_libs,
    shared_libs_auto_upgrade_disabled,
)
from pipx.util import (
    PipxError,
    exec_app,
    full_package_description,
    get_site_packages,
    get_venv_paths,
    pipx_wrap,
    rmdir,
    run_subprocess,
    subprocess_post_check,
)
from pipx.venv_inspect import VenvMetadata, inspect_venv

_LOGGER: Final[logging.Logger] = logging.getLogger(__name__)

# Keyed on full path so global vs user-local venvs with the same name don't
# collide; deduped per-session because ``upgrade-all --backend uv`` against
# many pip-backed venvs would otherwise repeat the multi-line warning per venv.
_BACKEND_OVERRIDE_WARNED: Final[set[str]] = set()


def reset_backend_override_warnings() -> None:
    # Test-only: production callers never reset this.
    _BACKEND_OVERRIDE_WARNED.clear()


def _resolve_backend_for_venv(
    *,
    root: Path,
    existing: bool,
    metadata_backend: str | None,
    cli_backend: str | None,
    env_backend: str | None,
) -> tuple[str, str]:
    """Apply cli > metadata > env > auto precedence, warning on conflict.

    Pulled out of ``Venv.__init__`` so the precedence rule is unit-testable
    without instantiating a real venv.
    """
    # Existing venvs lock to their recorded backend: ``uv pip install`` against
    # a pip-backed venv leaves it in an inconsistent state. ``pipx reinstall
    # NAME --backend X`` is the supported flip.
    if existing and cli_backend is not None and cli_backend != metadata_backend:
        warning_key = str(root)
        if warning_key not in _BACKEND_OVERRIDE_WARNED:
            _BACKEND_OVERRIDE_WARNED.add(warning_key)
            _LOGGER.warning(
                pipx_wrap(
                    f"""
                    {hazard}  Ignoring --backend={cli_backend} for existing venv {root.name!r}
                    (recorded backend is {metadata_backend!r}). Run
                    `pipx reinstall {root.name} --backend {cli_backend}` to flip the venv.
                    """,
                    subsequent_indent=" " * 4,
                )
            )
        cli_backend = None
    # Direct ``Venv(...)`` callers (including unit tests) won't have threaded
    # ``env_backend`` through; honor ``PIPX_DEFAULT_BACKEND`` so auto-detect
    # doesn't fire when the operator already opted in via env.
    if env_backend is None:
        env_backend = env_default_backend()
    return resolve_backend_name(cli_value=cli_backend, env_value=env_backend, metadata_value=metadata_backend)


class VenvContainer:
    """A collection of venvs managed by pipx."""

    def __init__(self, root: Path):
        self._root = root

    def __repr__(self) -> str:
        return f"VenvContainer({str(self._root)!r})"

    def __str__(self) -> str:
        return str(self._root)

    def iter_venv_dirs(self) -> Generator[Path, None, None]:
        """Iterate venv directories in this container."""
        if not self._root.is_dir():
            return
        for entry in self._root.iterdir():
            if not entry.is_dir():
                continue
            yield entry

    def get_venv_dir(self, package_name: str) -> Path:
        """Return the expected venv path for given `package_name`."""
        return self._root.joinpath(canonicalize_name(package_name))


class Venv:
    """Abstraction for a virtual environment with various useful methods for pipx"""

    def __init__(
        self,
        path: Path,
        *,
        verbose: bool = False,
        python: str = DEFAULT_PYTHON,
        backend: str | None = None,
        env_backend: str | None = None,
    ) -> None:
        self.root = path
        self.python = python
        self.bin_path, self.python_path, self.man_path = get_venv_paths(self.root)
        self.pipx_metadata = PipxMetadata(venv_dir=path)
        self.verbose = verbose
        self.do_animation = not verbose
        try:
            self._existing = self.root.exists() and bool(next(self.root.iterdir()))
        except StopIteration:
            self._existing = False
        self._backend_name, self._backend_source = _resolve_backend_for_venv(
            root=self.root,
            existing=self._existing,
            metadata_backend=self.pipx_metadata.backend if self._existing else None,
            cli_backend=backend,
            env_backend=env_backend,
        )
        self._backend: Backend | None = None
        self._uses_shared_libs_cache: bool | None = None

    @property
    def backend(self) -> Backend:
        if self._backend is None:
            self._backend = get_backend(self._backend_name)
        return self._backend

    @property
    def backend_name(self) -> str:
        return self._backend_name

    @property
    def backend_source(self) -> str:
        return self._backend_source

    def check_upgrade_shared_libs(self, verbose: bool, pip_args: list[str], force_upgrade: bool = False):
        """
        If necessary, run maintenance tasks to keep the shared libs up-to-date.

        This can trigger `pip install`/`pip install --upgrade` operations,
        so we expect the caller to provide sensible `pip_args`
        ( provided by the user in the current CLI call
        or retrieved from the metadata of a previous installation)
        """
        if self._existing and self.uses_shared_libs:
            if shared_libs.is_valid:
                if force_upgrade:
                    shared_libs.upgrade(verbose=verbose, pip_args=pip_args)
                elif shared_libs_auto_upgrade_disabled():
                    _LOGGER.info(f"Skipping shared libs auto-upgrade because {DISABLE_SHARED_LIBS_AUTO_UPGRADE} is set.")
                elif shared_libs.needs_upgrade:
                    shared_libs.upgrade(verbose=verbose, pip_args=pip_args)
            else:
                shared_libs.create(verbose=verbose, pip_args=pip_args)

            if not shared_libs.is_valid:
                raise PipxError(
                    pipx_wrap(
                        f"""
                        Error: pipx's shared venv {shared_libs.root} is invalid
                        and needs re-installation. To fix this, install or
                        reinstall a package. For example:
                        """
                    )
                    + f"\n  pipx install {self.root.name} --force",
                    wrap_message=False,
                )

    @property
    def name(self) -> str:
        if self.pipx_metadata.main_package.package is not None:
            venv_name = f"{self.pipx_metadata.main_package.package}{self.pipx_metadata.main_package.suffix}"
        else:
            venv_name = self.root.name
        return venv_name

    @property
    def uses_shared_libs(self) -> bool:
        if not self._existing:
            return self.backend.needs_shared_libs()
        if self._uses_shared_libs_cache is not None:
            return self._uses_shared_libs_cache
        # Metadata 0.6+ records ``backend`` authoritatively; older recordings
        # (or missing file) fall back to the .pth probe so legacy venvs still
        # report correctly.
        recorded_version = self.pipx_metadata.read_metadata_version
        if recorded_version is not None and recorded_version >= "0.6":
            answer = self.pipx_metadata.backend == "pip"
        else:
            answer = next(self.root.glob(f"**/{PIPX_SHARED_PTH}"), None) is not None
        self._uses_shared_libs_cache = answer
        return answer

    @property
    def package_metadata(self) -> dict[str, PackageInfo]:
        return_dict = self.pipx_metadata.injected_packages.copy()
        if self.pipx_metadata.main_package.package is not None:
            return_dict[self.pipx_metadata.main_package.package] = self.pipx_metadata.main_package
        return return_dict

    @property
    def main_package_name(self) -> str:
        if self.pipx_metadata.main_package.package is None:
            # This is OK, because if no metadata, we are pipx < v0.15.0.0 and
            #   venv_name==main_package_name
            return self.root.name
        else:
            return self.pipx_metadata.main_package.package

    def create_venv(self, venv_args: list[str], pip_args: list[str], override_shared: bool = False) -> None:
        """
        override_shared -- Override installing shared libraries to the pipx shared directory (default False)
        """
        _LOGGER.info("Creating virtual environment")
        if override_shared:
            assert_not_pip_under_uv("pip", self._backend_name)
        self.backend.create_venv(
            self.root,
            python=self.python,
            venv_args=venv_args,
            pip_args=pip_args,
            include_pip=override_shared,
            verbose=self.verbose,
        )

        self.pipx_metadata.venv_args = venv_args
        # Persist the chosen backend on disk only when actually creating the venv.
        # Runtime overrides on existing venvs (e.g. `--backend uv pipx upgrade ...`)
        # must not silently flip the recorded backend.
        self.pipx_metadata.backend = self._backend_name
        self.pipx_metadata.python_version = self.get_python_version()
        source_interpreter = shutil.which(self.python)
        if source_interpreter:
            self.pipx_metadata.source_interpreter = Path(source_interpreter)

    def safe_to_remove(self) -> bool:
        return not self._existing

    def remove_venv(self) -> None:
        if self.safe_to_remove():
            rmdir(self.root)
        else:
            _LOGGER.warning(
                pipx_wrap(
                    f"""
                    {hazard}  Not removing existing venv {self.root} because it
                    was not created in this session
                    """,
                    subsequent_indent=" " * 4,
                )
            )

    def upgrade_packaging_libraries(self, pip_args: list[str]) -> None:
        if self.uses_shared_libs:
            shared_libs.upgrade(pip_args=pip_args, verbose=self.verbose)
            return
        # Short-circuit before instantiating UvBackend so ``upgrade-all`` over
        # uv venvs still works after uv has been uninstalled (the no-op upgrade
        # would have happened anyway).
        if self._backend_name == "uv":
            return
        self.backend.upgrade_packaging_libraries(self.python_path, pip_args, verbose=self.verbose)

    def uninstall_package(self, package: str, was_injected: bool = False):
        try:
            _LOGGER.info("Uninstalling %s", package)
            with animate(f"uninstalling {package}", self.do_animation):
                self.backend.uninstall(
                    venv_root=self.root,
                    venv_python=self.python_path,
                    package=package,
                    verbose=self.verbose,
                )
        except PipxError as e:
            _LOGGER.info(e)
            raise PipxError(f"Error uninstalling {package}.") from None

        if was_injected:
            self.pipx_metadata.injected_packages.pop(package)
            self.pipx_metadata.write()

    def install_package(
        self,
        package_name: str,
        package_or_url: str,
        pip_args: list[str],
        include_dependencies: bool,
        include_apps: bool,
        is_main_package: bool,
        suffix: str = "",
        install_only_pip_args: list[str] | None = None,
    ) -> None:
        # package_name in package specifier can mismatch URL due to user error
        package_or_url = fix_package_name(package_or_url, package_name)

        # check syntax and clean up spec and pip_args
        (package_or_url, pip_args) = parse_specifier_for_install(package_or_url, pip_args)
        install_pip_args = [*(install_only_pip_args or []), *pip_args]

        _LOGGER.info("Installing %s", package_descr := full_package_description(package_name, package_or_url))
        with animate(f"installing {package_descr}", self.do_animation):
            process = self.backend.install(
                venv_root=self.root,
                venv_python=self.python_path,
                requirements=[package_or_url],
                pip_args=install_pip_args,
                verbose=self.verbose,
            )
        if process.returncode:
            raise PipxError(f"Error installing {full_package_description(package_name, package_or_url)}.")

        self.update_package_metadata(
            package_name=package_name,
            package_or_url=package_or_url,
            pip_args=pip_args,
            include_dependencies=include_dependencies,
            include_apps=include_apps,
            is_main_package=is_main_package,
            suffix=suffix,
        )

        # Verify package installed ok
        if self.package_metadata[package_name].package_version is None:
            raise PipxError(
                f"Unable to install "
                f"{full_package_description(package_name, package_or_url)}.\n"
                f"Check the name or spec for errors, and verify that it can "
                f"be installed with pip.",
                wrap_message=False,
            )

    def install_unmanaged_packages(self, requirements: list[str], pip_args: list[str]) -> None:
        """Install packages in the venv, but do not record them in the metadata."""
        _LOGGER.info("Installing %s", package_descr := ", ".join(requirements))
        with animate(f"installing {package_descr}", self.do_animation):
            process = self.backend.install(
                venv_root=self.root,
                venv_python=self.python_path,
                requirements=list(requirements),
                pip_args=pip_args,
                verbose=self.verbose,
            )
        if process.returncode:
            raise PipxError(f"Error installing {', '.join(requirements)}.")

    def install_package_no_deps(self, package_or_url: str, pip_args: list[str]) -> str:
        with animate(f"determining package name from {package_or_url!r}", self.do_animation):
            old_package_set = self.list_installed_packages()
            process = self.backend.install(
                venv_root=self.root,
                venv_python=self.python_path,
                requirements=[package_or_url],
                pip_args=pip_args,
                no_deps=True,
                log_pip_errors=False,
                verbose=self.verbose,
            )
        if process.returncode:
            error_output = (process.stderr or process.stdout or "").strip()
            if error_output:
                raise PipxError(error_output, wrap_message=False)
            raise PipxError(
                f"""
                Error determining package name from spec {package_or_url!r}.
                See installer output for details.
                """
            )

        installed_packages = self.list_installed_packages() - old_package_set
        if len(installed_packages) == 1:
            package_name = installed_packages.pop()
            _LOGGER.info(f"Determined package name: {package_name}")
        else:
            _LOGGER.info(f"old_package_set = {old_package_set}")
            _LOGGER.info(f"install_packages = {installed_packages}")
            raise PipxError(
                f"""
                Cannot determine package name from spec {package_or_url!r}.
                Check package spec for errors.
                """
            )

        return package_name

    def get_venv_metadata_for_package(self, package_name: str, package_extras: set[str]) -> VenvMetadata:
        data_start = time.time()
        venv_metadata = inspect_venv(package_name, package_extras, self.bin_path, self.python_path, self.man_path)
        _LOGGER.info(f"get_venv_metadata_for_package: {1e3 * (time.time() - data_start):.0f}ms")
        return venv_metadata

    def update_package_metadata(
        self,
        package_name: str,
        package_or_url: str,
        pip_args: list[str],
        include_dependencies: bool,
        include_apps: bool,
        is_main_package: bool,
        suffix: str = "",
        pinned: bool = False,
    ) -> None:
        venv_package_metadata = self.get_venv_metadata_for_package(package_name, get_extras(package_or_url))
        package_info = PackageInfo(
            package=package_name,
            package_or_url=parse_specifier_for_metadata(package_or_url),
            pip_args=pip_args,
            include_apps=include_apps,
            include_dependencies=include_dependencies,
            apps=venv_package_metadata.apps,
            app_paths=venv_package_metadata.app_paths,
            apps_of_dependencies=venv_package_metadata.apps_of_dependencies,
            app_paths_of_dependencies=venv_package_metadata.app_paths_of_dependencies,
            man_pages=venv_package_metadata.man_pages,
            man_paths=venv_package_metadata.man_paths,
            man_pages_of_dependencies=venv_package_metadata.man_pages_of_dependencies,
            man_paths_of_dependencies=venv_package_metadata.man_paths_of_dependencies,
            package_version=venv_package_metadata.package_version,
            suffix=suffix,
            pinned=pinned,
        )
        if is_main_package:
            self.pipx_metadata.main_package = package_info
        else:
            self.pipx_metadata.injected_packages[package_name] = package_info

        self.pipx_metadata.write()

    def get_python_version(self) -> str:
        return run_subprocess([str(self.python_path), "--version"]).stdout.strip()

    def list_installed_packages(self, not_required: bool = False) -> set[str]:
        return self.backend.list_installed(
            venv_root=self.root,
            venv_python=self.python_path,
            not_required=not_required,
        )

    def _find_entry_point(self, app: str) -> EntryPoint | None:
        if not self.python_path.exists():
            return None
        dists = Distribution.discover(name=self.main_package_name, path=[str(get_site_packages(self.python_path))])
        for dist in dists:
            for ep in dist.entry_points:
                if ep.group == "pipx.run":
                    if ep.name == app:
                        return ep
                    # Try to infer app name from dist's metadata if given
                    # local path
                    if Path(app).exists() and dist.metadata["Name"] == ep.name:
                        return ep
        return None

    def run_app(self, app: str, filename: str, app_args: list[str]) -> NoReturn:
        entry_point = self._find_entry_point(app)

        # No [pipx.run] entry point; default to run console script.
        if entry_point is None:
            exec_app([str(self.bin_path / filename)] + app_args)

        # Evaluate and execute the entry point.
        _LOGGER.info("Using discovered entry point for 'pipx run'")
        module, attr = entry_point.module, entry_point.attr
        code = f"import sys, {module}\nsys.argv[0] = {entry_point.name!r}\nsys.exit({module}.{attr}())\n"
        exec_app([str(self.python_path), "-c", code] + app_args)

    def has_app(self, app: str, filename: str) -> bool:
        if self._find_entry_point(app) is not None:
            return True
        return (self.bin_path / filename).is_file()

    def has_package(self, package_name: str) -> bool:
        return bool(list(Distribution.discover(name=package_name, path=[str(get_site_packages(self.python_path))])))

    def upgrade_package_no_metadata(self, package_name: str, pip_args: list[str]) -> None:
        _LOGGER.info("Upgrading %s", package_descr := full_package_description(package_name, package_name))
        with animate(f"upgrading {package_descr}", self.do_animation):
            process = self.backend.install(
                venv_root=self.root,
                venv_python=self.python_path,
                requirements=[package_name],
                pip_args=pip_args,
                upgrade=True,
                log_pip_errors=False,
                verbose=self.verbose,
            )
        subprocess_post_check(process)

    def upgrade_package(
        self,
        package_name: str,
        package_or_url: str,
        pip_args: list[str],
        include_dependencies: bool,
        include_apps: bool,
        is_main_package: bool,
        suffix: str = "",
    ) -> None:
        _LOGGER.info("Upgrading %s", package_descr := full_package_description(package_name, package_or_url))
        with animate(f"upgrading {package_descr}", self.do_animation):
            process = self.backend.install(
                venv_root=self.root,
                venv_python=self.python_path,
                requirements=[package_or_url],
                pip_args=pip_args,
                upgrade=True,
                log_pip_errors=False,
                verbose=self.verbose,
            )
        subprocess_post_check(process)

        self.update_package_metadata(
            package_name=package_name,
            package_or_url=package_or_url,
            pip_args=pip_args,
            include_dependencies=include_dependencies,
            include_apps=include_apps,
            is_main_package=is_main_package,
            suffix=suffix,
        )

    def _run_pip(self, cmd: list[str]) -> "CompletedProcess[str]":
        return self.backend.run_raw_pip(
            venv_root=self.root,
            venv_python=self.python_path,
            args=cmd,
            verbose=self.verbose,
        )

    def run_pip_get_exit_code(self, cmd: list[str]) -> ExitCode:
        process = self.backend.run_raw_pip(
            venv_root=self.root,
            venv_python=self.python_path,
            args=cmd,
            capture_stdout=False,
            capture_stderr=False,
            verbose=self.verbose,
        )
        if process.returncode:
            cmd_str = " ".join(str(c) for c in process.args)
            _LOGGER.error(f"{cmd_str!r} failed")
        return ExitCode(process.returncode)


__all__ = [
    "Venv",
    "VenvContainer",
    "reset_backend_override_warnings",
]



import datetime
import hashlib
import logging
import re
import sys
import time
import urllib.parse
import urllib.request
from pathlib import Path
from shutil import which
from typing import Final, NoReturn

from packaging.requirements import InvalidRequirement, Requirement

from pipx import paths
from pipx.backends import UV, resolve_backend_name
from pipx.commands.common import package_name_from_spec
from pipx.commands.inject import inject_dep
from pipx.commands.run_uv import run_script_via_uv_run, run_via_uv_tool_run
from pipx.constants import TEMP_VENV_EXPIRATION_THRESHOLD_DAYS, WINDOWS
from pipx.emojis import hazard
from pipx.util import (
    PipxError,
    exec_app,
    get_pypackage_bin_path,
    pipx_wrap,
    rmdir,
    run_pypackage_bin,
)
from pipx.venv import Venv

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

_LOGGER: Final[logging.Logger] = logging.getLogger(__name__)

_VENV_EXPIRED_FILENAME: Final[str] = "pipx_expired_venv"

_APP_NOT_FOUND_ERROR_MESSAGE: Final[str] = """\
'{app}' executable script not found in package '{package_name}'.
Available executable scripts:
    {app_lines}"""


def maybe_script_content(app: str, is_path: bool) -> str | Path | None:
    """If the app is a script, return its content.
    Return None if it should be treated as a package name."""

    # Look for a local file first.
    app_path = Path(app)
    if app_path.is_file():
        return app_path
    # In case it's a named pipe, read it out to pass to the interpreter
    if app_path.is_fifo():
        return app_path.read_text(encoding="utf-8")

    if is_path:
        raise PipxError(f"The specified path {app} does not exist")

    # Check for a URL
    if urllib.parse.urlparse(app).scheme:
        if not app.endswith(".py"):
            raise PipxError(
                """
                pipx will only execute apps from the internet directly if they
                end with '.py'. To run from an SVN, try pipx --spec URL BINARY
                """
            )
        _LOGGER.info("Detected url. Downloading and executing as a Python file.")

        return _http_get_request(app)

    # Otherwise, it's a package
    return None


def run_script(
    content: str | Path,
    app_args: list[str],
    python: str,
    pip_args: list[str],
    venv_args: list[str],
    verbose: bool,
    use_cache: bool,
    *,
    backend: str | None = None,
    env_backend: str | None = None,
    resolved_backend: str | None = None,
    script_source: Path | None = None,
    dependencies: list[str] | None = None,
) -> NoReturn:
    requirements = _get_requirements_from_script(content)

    if dependencies and not requirements:
        # Plain scripts have nowhere to record extra requirements; the pip path
        # silently dropped ``--with`` here, but a clear error is better.
        raise PipxError(
            "--with packages can only be applied to scripts with PEP 723 inline metadata "
            "(`# /// script` block). Add the dependencies to the script's metadata or run "
            "via `pipx run --spec`."
        )

    if resolved_backend == UV and requirements is not None:
        if script_source is not None:
            run_script_via_uv_run(
                script_path=script_source,
                app_args=app_args,
                python=python,
                pip_args=pip_args,
                venv_args=venv_args,
                use_cache=use_cache,
                verbose=verbose,
                dependencies=dependencies,
            )
        # URL / named-pipe content has no on-disk path for ``uv run --script``;
        # warn so users on ``--backend uv`` notice they lose uv's cache and
        # Python-managing semantics for this run.
        _LOGGER.warning(
            pipx_wrap(
                f"""
                {hazard}  Script content is not a local file; building a
                pipx-managed venv via the uv backend instead of `uv run
                --script`. The uv cache and uv-managed Python features
                will not apply to this run.
                """,
                subsequent_indent=" " * 4,
            )
        )

    if not requirements:
        python_path = Path(python)
    else:
        # Note that the environment name is based on the identified
        # requirements, and *not* on the script name. This is deliberate, as
        # it ensures that two scripts with the same requirements can use the
        # same environment, which means fewer environments need to be
        # managed. The requirements are normalised (in
        # _get_requirements_from_script), so that irrelevant differences in
        # whitespace, and similar, don't prevent environment sharing.
        venv_dir = _get_temporary_venv_path(requirements, python, pip_args, venv_args, resolved_backend or "pip")
        venv = Venv(venv_dir, backend=backend, env_backend=env_backend)
        _prepare_venv_cache(venv, None, use_cache)
        if venv_dir.exists():
            _LOGGER.info(f"Reusing cached venv {venv_dir}")
        else:
            venv = Venv(venv_dir, python=python, verbose=verbose, backend=backend, env_backend=env_backend)
            venv.check_upgrade_shared_libs(pip_args=pip_args, verbose=verbose)
            venv.create_venv(venv_args, pip_args)
            try:
                venv.install_unmanaged_packages(requirements, pip_args)
            except:
                # Package installation failed, so mark the cache as expired.
                # This ensures an attempt is made to re-install requirements
                # when `pipx run` is next executed, rather than just failing.
                (venv_dir / _VENV_EXPIRED_FILENAME).touch()
                raise
        python_path = venv.python_path

    if isinstance(content, Path):
        exec_app([python_path, content, *app_args])
    else:
        exec_app([python_path, "-c", content, *app_args])


def run_package(
    app: str,
    package_or_url: str,
    dependencies: list[str],
    app_args: list[str],
    python: str,
    pip_args: list[str],
    venv_args: list[str],
    pypackages: bool,
    verbose: bool,
    use_cache: bool,
    *,
    backend: str | None = None,
    env_backend: str | None = None,
    resolved_backend: str | None = None,
    no_path_check: bool = False,
) -> NoReturn:
    if not no_path_check and which(app):
        _LOGGER.warning(
            pipx_wrap(
                f"""
                {hazard}  {app} is already on your PATH and installed at
                {which(app)}. Downloading and running anyway.
                """,
                subsequent_indent=" " * 4,
            )
        )

    if WINDOWS:
        app_filename = f"{app}.exe"
        _LOGGER.info(f"Assuming app is {app_filename!r} (Windows only)")
    else:
        app_filename = app

    pypackage_bin_path = get_pypackage_bin_path(app)
    if pypackage_bin_path.exists():
        _LOGGER.info(f"Using app in local __pypackages__ directory at '{pypackage_bin_path}'")
        run_pypackage_bin(pypackage_bin_path, app_args)
    if pypackages:
        raise PipxError(
            f"""
            '--pypackages' flag was passed, but '{pypackage_bin_path}' was
            not found. See https://github.com/cs01/pythonloc to learn how to
            install here, or omit the flag.
            """
        )

    venv_dir = _get_temporary_venv_path([package_or_url], python, pip_args, venv_args, resolved_backend or "pip")

    venv = Venv(venv_dir, backend=backend, env_backend=env_backend)
    bin_path = venv.bin_path / app_filename
    _prepare_venv_cache(venv, bin_path, use_cache)

    if venv.has_app(app, app_filename):
        _LOGGER.info(f"Reusing cached venv {venv_dir}")
    else:
        _LOGGER.info(f"venv location is {venv_dir}")
        venv, app, app_filename = _prepare_venv(
            Path(venv_dir),
            package_or_url,
            app,
            app_filename,
            python,
            pip_args,
            venv_args,
            use_cache,
            verbose,
            backend=backend,
            env_backend=env_backend,
        )

    for dependency in dependencies:
        inject_dep(
            venv_dir=venv_dir,
            package_name=None,
            package_spec=dependency,
            pip_args=pip_args,
            verbose=verbose,
            include_apps=False,
            include_dependencies=False,
            force=False,
            backend=backend,
            env_backend=env_backend,
        )
    venv.run_app(app, app_filename, app_args)


def run(
    app: str,
    spec: str | None,
    dependencies: list[str],
    is_path: bool,
    app_args: list[str],
    python: str,
    pip_args: list[str],
    venv_args: list[str],
    pypackages: bool,
    verbose: bool,
    use_cache: bool,
    *,
    no_path_check: bool = False,
    backend: str | None = None,
    env_backend: str | None = None,
) -> NoReturn:
    """Installs venv to temporary dir (or reuses cache), then runs app from
    package
    """

    # For any package, we need to just use the name
    try:
        package_name = Requirement(app).name
    except InvalidRequirement:
        # Raw URLs to scripts are supported, too, so continue if
        # we can't parse this as a package
        package_name = app

    # ``resolved_backend`` only decides ROUTING (uv tool run vs Venv); cli/env
    # stay separate when we hand off so the Venv's source-attribution stays right.
    resolved_backend, _ = resolve_backend_name(cli_value=backend, env_value=env_backend)
    use_uvx = resolved_backend == UV and not pypackages

    content = None if spec is not None else maybe_script_content(app, is_path)
    if content is not None:
        run_script(
            content,
            app_args,
            python,
            pip_args,
            venv_args,
            verbose,
            use_cache,
            backend=backend,
            env_backend=env_backend,
            resolved_backend=resolved_backend,
            script_source=Path(app) if isinstance(content, Path) else None,
            dependencies=dependencies,
        )
    elif use_uvx:
        run_via_uv_tool_run(
            app=app,
            package_or_url=spec if spec is not None else app,
            dependencies=dependencies,
            app_args=app_args,
            python=python,
            pip_args=pip_args,
            venv_args=venv_args,
            use_cache=use_cache,
            verbose=verbose,
            no_path_check=no_path_check,
        )
    else:
        package_or_url = spec if spec is not None else app
        run_package(
            package_name,
            package_or_url,
            dependencies,
            app_args,
            python,
            pip_args,
            venv_args,
            pypackages,
            verbose,
            use_cache,
            backend=backend,
            env_backend=env_backend,
            resolved_backend=resolved_backend,
            no_path_check=no_path_check,
        )


def _prepare_venv(
    venv_dir: Path,
    package_or_url: str,
    app: str,
    app_filename: str,
    python: str,
    pip_args: list[str],
    venv_args: list[str],
    use_cache: bool,
    verbose: bool,
    *,
    backend: str | None = None,
    env_backend: str | None = None,
) -> tuple[Venv, str, str]:
    venv = Venv(venv_dir, python=python, verbose=verbose, backend=backend, env_backend=env_backend)
    venv.check_upgrade_shared_libs(pip_args=pip_args, verbose=verbose)

    if venv.pipx_metadata.main_package.package is not None:
        package_name = venv.pipx_metadata.main_package.package
    else:
        package_name = package_name_from_spec(package_or_url, python, pip_args=pip_args, verbose=verbose)

    override_shared = package_name == "pip"

    venv.create_venv(venv_args, pip_args, override_shared)
    venv.install_package(
        package_name=package_name,
        package_or_url=package_or_url,
        pip_args=pip_args,
        include_dependencies=False,
        include_apps=True,
        is_main_package=True,
    )

    if not venv.has_app(app, app_filename):
        apps = venv.pipx_metadata.main_package.apps

        # If there's a single app inside the package, run that by default
        if app == package_name and len(apps) == 1:
            app = apps[0]
            print(f"NOTE: running app {app!r} from {package_name!r}")
            if WINDOWS:
                app_filename = f"{app}.exe"
                _LOGGER.info(f"Assuming app is {app_filename!r} (Windows only)")
            else:
                app_filename = app
        else:
            all_apps = (f"{a} - usage: 'pipx run --spec {package_or_url} {a} [arguments?]'" for a in apps)
            raise PipxError(
                _APP_NOT_FOUND_ERROR_MESSAGE.format(
                    app=app,
                    package_name=package_name,
                    app_lines="\n    ".join(all_apps),
                ),
                wrap_message=False,
            )

    if not use_cache:
        # Let future _remove_all_expired_venvs know to remove this
        (venv_dir / _VENV_EXPIRED_FILENAME).touch()

    return venv, app, app_filename


def _get_temporary_venv_path(
    requirements: list[str],
    python: str,
    pip_args: list[str],
    venv_args: list[str],
    backend: str,
) -> Path:
    """Hash venv-affecting inputs to a deterministic cache path.

    ``backend`` is part of the key so pip- and uv-backed temp venvs for the
    same package coexist instead of stomping on each other.
    """
    digest = hashlib.sha256()
    digest.update("".join(requirements).encode())
    digest.update(python.encode())
    digest.update("".join(pip_args).encode())
    digest.update("".join(venv_args).encode())
    digest.update(backend.encode())
    venv_folder_name = digest.hexdigest()[:15]  # 15 chosen arbitrarily
    return Path(paths.ctx.venv_cache) / venv_folder_name


def _is_temporary_venv_expired(venv_dir: Path) -> bool:
    created_time_sec = venv_dir.stat().st_ctime
    current_time_sec = time.mktime(datetime.datetime.now().timetuple())
    age = current_time_sec - created_time_sec
    expiration_threshold_sec = 60 * 60 * 24 * TEMP_VENV_EXPIRATION_THRESHOLD_DAYS
    return age > expiration_threshold_sec or (venv_dir / _VENV_EXPIRED_FILENAME).exists()


def _prepare_venv_cache(venv: Venv, bin_path: Path | None, use_cache: bool) -> None:
    venv_dir = venv.root
    if not use_cache and (bin_path is None or bin_path.exists()):
        _LOGGER.info(f"Removing cached venv {venv_dir!s}")
        rmdir(venv_dir)
    _remove_all_expired_venvs()


def _remove_all_expired_venvs() -> None:
    for venv_dir in Path(paths.ctx.venv_cache).iterdir():
        if _is_temporary_venv_expired(venv_dir):
            _LOGGER.info(f"Removing expired venv {venv_dir!s}")
            rmdir(venv_dir)


def _http_get_request(url: str) -> str:
    try:
        res = urllib.request.urlopen(url)
        charset = res.headers.get_content_charset() or "utf-8"
        return res.read().decode(charset)
    except Exception as e:
        _LOGGER.debug("Uncaught Exception:", exc_info=True)
        raise PipxError(str(e)) from e


# Pattern from PEP 723 / inline script metadata spec.
_INLINE_SCRIPT_METADATA: Final[re.Pattern[str]] = re.compile(
    r"""
    ^\#\ ///\ (?P<type>[a-zA-Z0-9-]+)$ \s   # opening fence: ``# /// <type>``
    (?P<content> (^\#(|\ .*)$ \s)+ )        # body: lines starting with ``#`` or ``# ``
    ^\#\ ///$                               # closing fence: ``# ///``
    """,
    re.VERBOSE | re.MULTILINE,
)


def _get_requirements_from_script(content: str | Path) -> list[str] | None:
    """
    Supports inline script metadata.
    """

    if isinstance(content, Path):
        content = content.read_text(encoding="utf-8")

    name = "script"

    # Windows is currently getting un-normalized line endings, so normalize
    content = content.replace("\r\n", "\n")

    matches = [m for m in _INLINE_SCRIPT_METADATA.finditer(content) if m.group("type") == name]

    if not matches:
        pyproject_matches = [m for m in _INLINE_SCRIPT_METADATA.finditer(content) if m.group("type") == "pyproject"]
        if pyproject_matches:
            _LOGGER.error(
                pipx_wrap(
                    f"""
                    {hazard}  Using old form of requirements table. Use updated PEP
                    723 syntax by replacing `# /// pyproject` with `# /// script`
                    and `run.dependencies` (or `run.requirements`) with
                    `dependencies`.
                    """,
                    subsequent_indent=" " * 4,
                )
            )
            raise ValueError("Old 'pyproject' table found")
        return None

    if len(matches) > 1:
        raise ValueError(f"Multiple {name} blocks found")

    content = "".join(
        line[2:] if line.startswith("# ") else line[1:] for line in matches[0].group("content").splitlines(keepends=True)
    )

    pyproject = tomllib.loads(content)

    requirements = []
    for requirement in pyproject.get("dependencies", []):
        try:
            parsed_requirement = Requirement(requirement)
        except InvalidRequirement as exc:
            raise PipxError(f"Invalid requirement {requirement}: {exc}") from exc

        # Use the normalised form of the requirement
        requirements.append(str(parsed_requirement))

    return requirements


__all__ = [
    "maybe_script_content",
    "run",
    "run_package",
    "run_script",
]



import json
import logging
import sys
from collections.abc import Collection
from pathlib import Path
from typing import Any

from pipx import paths
from pipx.colors import bold
from pipx.commands.common import VenvProblems, get_venv_summary, venv_health_check
from pipx.constants import EXIT_CODE_LIST_PROBLEM, EXIT_CODE_OK, ExitCode
from pipx.emojis import sleep
from pipx.pipx_metadata_file import JsonEncoderHandlesPath, PipxMetadata
from pipx.venv import Venv, VenvContainer

logger = logging.getLogger(__name__)

PIPX_SPEC_VERSION = "0.1"


def get_venv_metadata_summary(venv_dir: Path) -> tuple[PipxMetadata, VenvProblems, str]:
    venv = Venv(venv_dir)

    (venv_problems, warning_message) = venv_health_check(venv)
    if venv_problems.any_():
        return (PipxMetadata(venv_dir, read=False), venv_problems, warning_message)

    return (venv.pipx_metadata, venv_problems, "")


def list_short(venv_dirs: Collection[Path]) -> VenvProblems:
    all_venv_problems = VenvProblems()
    for venv_dir in venv_dirs:
        venv_metadata, venv_problems, warning_str = get_venv_metadata_summary(venv_dir)
        if venv_problems.any_():
            logger.warning(warning_str)
        else:
            print(
                venv_metadata.main_package.package,
                venv_metadata.main_package.package_version,
            )
        all_venv_problems.or_(venv_problems)

    return all_venv_problems


def list_text(venv_dirs: Collection[Path], include_injected: bool, venv_root_dir: str) -> VenvProblems:
    print(f"venvs are in {bold(venv_root_dir)}")
    print(f"apps are exposed on your $PATH at {bold(str(paths.ctx.bin_dir))}")
    print(f"manual pages are exposed at {bold(str(paths.ctx.man_dir))}")

    all_venv_problems = VenvProblems()
    for venv_dir in venv_dirs:
        package_summary, venv_problems = get_venv_summary(venv_dir, include_injected=include_injected)
        if venv_problems.any_():
            logger.warning(package_summary)
        else:
            print(package_summary)
        all_venv_problems.or_(venv_problems)

    return all_venv_problems


def list_json(venv_dirs: Collection[Path]) -> VenvProblems:
    warning_messages = []
    spec_metadata: dict[str, Any] = {
        "pipx_spec_version": PIPX_SPEC_VERSION,
        "venvs": {},
    }
    all_venv_problems = VenvProblems()
    for venv_dir in venv_dirs:
        (venv_metadata, venv_problems, warning_str) = get_venv_metadata_summary(venv_dir)
        all_venv_problems.or_(venv_problems)
        if venv_problems.any_():
            warning_messages.append(warning_str)
            continue

        spec_metadata["venvs"][venv_dir.name] = {}
        spec_metadata["venvs"][venv_dir.name]["metadata"] = venv_metadata.to_dict()

    print(json.dumps(spec_metadata, indent=4, sort_keys=True, cls=JsonEncoderHandlesPath))
    for warning_message in warning_messages:
        logger.warning(warning_message)

    return all_venv_problems


def list_pinned(venv_dirs: Collection[Path], include_injected: bool) -> VenvProblems:
    all_venv_problems = VenvProblems()
    for venv_dir in venv_dirs:
        venv_metadata, venv_problems, warning_str = get_venv_metadata_summary(venv_dir)
        if venv_problems.any_():
            logger.warning(warning_str)
        else:
            if venv_metadata.main_package.pinned:
                print(
                    venv_metadata.main_package.package,
                    venv_metadata.main_package.package_version,
                )
            if include_injected:
                for pkg, info in venv_metadata.injected_packages.items():
                    if info.pinned:
                        print(pkg, info.package_version, f"(injected in venv {venv_dir.name})")
        all_venv_problems.or_(venv_problems)

    return all_venv_problems


def list_packages(
    venv_container: VenvContainer,
    include_injected: bool,
    json_format: bool,
    short_format: bool,
    pinned_only: bool,
) -> ExitCode:
    """Returns pipx exit code."""
    venv_dirs: Collection[Path] = sorted(venv_container.iter_venv_dirs())
    if not venv_dirs:
        print(f"nothing has been installed with pipx {sleep}", file=sys.stderr)

    if json_format:
        all_venv_problems = list_json(venv_dirs)
    elif short_format:
        all_venv_problems = list_short(venv_dirs)
    elif pinned_only:
        all_venv_problems = list_pinned(venv_dirs, include_injected)
    else:
        if not venv_dirs:
            return EXIT_CODE_OK
        all_venv_problems = list_text(venv_dirs, include_injected, str(venv_container))

    if all_venv_problems.bad_venv_name:
        logger.warning(
            "\nOne or more packages contain out-of-date internal data installed from a\n"
            "previous pipx version and need to be updated.\n"
            "    To fix, execute: pipx reinstall-all"
        )
    if all_venv_problems.invalid_interpreter:
        logger.warning(
            "\nOne or more packages have a missing python interpreter.\n    To fix, execute: pipx reinstall-all"
        )
    if all_venv_problems.missing_metadata:
        logger.warning(
            "\nOne or more packages have a missing internal pipx metadata.\n"
            "   They were likely installed using a pipx version before 0.15.0.0.\n"
            "   Please uninstall and install these package(s) to fix."
        )
    if all_venv_problems.not_installed:
        logger.warning(
            "\nOne or more packages are not installed properly.\n"
            "   Please uninstall and install these package(s) to fix."
        )

    if all_venv_problems.any_():
        print("", file=sys.stderr)
        return EXIT_CODE_LIST_PROBLEM

    return EXIT_CODE_OK



import json
import logging
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Any, Final, TypedDict

from pipx.backends._base import KNOWN_BACKENDS
from pipx.emojis import hazard
from pipx.util import PipxError, pipx_wrap

_LOGGER: Final[logging.Logger] = logging.getLogger(__name__)


PIPX_INFO_FILENAME = "pipx_metadata.json"


class _RawPackageInfo(TypedDict, total=False):
    """JSON-on-disk shape for :class:`PackageInfo`.

    ``total=False`` because the legacy-metadata migration runs against dumps
    from older pipx versions that lacked some keys.
    """

    package: str | None
    package_or_url: str | None
    pip_args: list[str]
    include_dependencies: bool
    include_apps: bool
    apps: list[str]
    app_paths: list[Path]
    apps_of_dependencies: list[str]
    app_paths_of_dependencies: dict[str, list[Path]]
    package_version: str
    man_pages: list[str]
    man_paths: list[Path]
    man_pages_of_dependencies: list[str]
    man_paths_of_dependencies: dict[str, list[Path]]
    suffix: str
    pinned: bool


class _RawMetadata(TypedDict, total=False):
    """JSON-on-disk shape for ``pipx_metadata.json``."""

    main_package: _RawPackageInfo
    python_version: str | None
    source_interpreter: Path | None
    venv_args: list[str]
    injected_packages: dict[str, _RawPackageInfo]
    backend: str
    pipx_metadata_version: str
    pinned: bool


class _RawSpecVenvEntry(TypedDict):
    metadata: _RawMetadata


class _RawSpecFile(TypedDict, total=False):
    # Wire format produced by ``pipx list --json`` and consumed by ``install-all``.
    venvs: dict[str, _RawSpecVenvEntry]
    pipx_spec_version: str


class JsonEncoderHandlesPath(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        # only handles what json.JSONEncoder doesn't understand by default
        if isinstance(obj, Path):
            return {"__type__": "Path", "__Path__": str(obj)}
        return super().default(obj)


def _json_decoder_object_hook(json_dict: dict[str, Any]) -> dict[str, Any] | Path:
    if json_dict.get("__type__") == "Path" and "__Path__" in json_dict:
        return Path(json_dict["__Path__"])
    return json_dict


@dataclass(frozen=True)
class PackageInfo:
    package: str | None
    package_or_url: str | None
    pip_args: list[str]
    include_dependencies: bool
    include_apps: bool
    apps: list[str]
    app_paths: list[Path]
    apps_of_dependencies: list[str]
    app_paths_of_dependencies: dict[str, list[Path]]
    package_version: str
    man_pages: list[str] = field(default_factory=list)
    man_paths: list[Path] = field(default_factory=list)
    man_pages_of_dependencies: list[str] = field(default_factory=list)
    man_paths_of_dependencies: dict[str, list[Path]] = field(default_factory=dict)
    suffix: str = ""
    pinned: bool = False


class PipxMetadata:
    # Only change this if file format changes
    # V0.1 -> original version
    # V0.2 -> Improve handling of suffixes
    # V0.3 -> Add man pages fields
    # V0.4 -> Add source interpreter
    # V0.5 -> Add pinned
    # V0.6 -> Add backend (pip|uv)
    __METADATA_VERSION__: str = "0.6"

    def __init__(self, venv_dir: Path, read: bool = True):
        self.venv_dir = venv_dir
        # Reasonable defaults for everything except the fields the caller
        # populates from the install spec (package, package_or_url, python_version).
        self.main_package = PackageInfo(
            package=None,
            package_or_url=None,
            pip_args=[],
            include_dependencies=False,
            include_apps=True,  # always True for main_package
            apps=[],
            app_paths=[],
            apps_of_dependencies=[],
            app_paths_of_dependencies={},
            man_pages=[],
            man_paths=[],
            man_pages_of_dependencies=[],
            man_paths_of_dependencies={},
            package_version="",
        )
        self.python_version: str | None = None
        self.source_interpreter: Path | None = None
        self.venv_args: list[str] = []
        self.injected_packages: dict[str, PackageInfo] = {}
        self.backend: str = "pip"
        # ``None`` until ``read()`` succeeds; lets callers tell a fresh
        # instance from one with authoritative on-disk values.
        self.read_metadata_version: str | None = None

        if read:
            self.read()

    def to_dict(self) -> dict[str, Any]:
        # Plain dict over _RawMetadata: asdict() returns dict[str, Any] and
        # consumers serialise straight to JSON.
        return {
            "main_package": asdict(self.main_package),
            "python_version": self.python_version,
            "source_interpreter": self.source_interpreter,
            "venv_args": self.venv_args,
            "injected_packages": {name: asdict(data) for (name, data) in self.injected_packages.items()},
            "backend": self.backend,
            "pipx_metadata_version": self.__METADATA_VERSION__,
        }

    def _convert_legacy_metadata(self, metadata_dict: _RawMetadata) -> _RawMetadata:
        version = metadata_dict["pipx_metadata_version"]
        if version in (self.__METADATA_VERSION__, "0.5"):
            pass
        elif version == "0.4":
            metadata_dict["pinned"] = False
        elif version in ("0.2", "0.3"):
            metadata_dict["source_interpreter"] = None
        elif version == "0.1":
            main_package_data = metadata_dict["main_package"]
            package_name = main_package_data["package"]
            if package_name is not None and package_name != self.venv_dir.name:
                # handle older suffixed packages gracefully
                main_package_data["suffix"] = self.venv_dir.name.replace(package_name, "")
            metadata_dict["source_interpreter"] = None
        else:
            raise PipxError(
                f"""
                {self.venv_dir.name}: Unknown metadata version {version}.
                Perhaps it was installed with a later version of pipx.
                """
            )
        # ``backend`` is absent from any pre-0.6 dump; default once here.
        metadata_dict.setdefault("backend", "pip")
        return metadata_dict

    def from_dict(self, input_dict: _RawMetadata) -> None:
        input_dict = self._convert_legacy_metadata(input_dict)
        self.main_package = PackageInfo(**input_dict["main_package"])
        self.python_version = input_dict.get("python_version")
        source_interpreter_raw = input_dict.get("source_interpreter")
        self.source_interpreter = Path(source_interpreter_raw) if source_interpreter_raw else None
        self.venv_args = input_dict.get("venv_args", [])
        self.injected_packages = {
            f"{name}{data.get('suffix', '')}": PackageInfo(**data)
            for (name, data) in input_dict.get("injected_packages", {}).items()
        }
        # Permissive: an unknown ``backend`` (manual edit, post-downgrade
        # dump from a newer pipx) falls back to pip with a warning so
        # ``pipx list`` / ``pipx uninstall`` keep working. Strict validation
        # still fires on write via :func:`pipx.backends.resolve_backend_name`.
        recorded_backend = input_dict.get("backend") or "pip"
        if recorded_backend not in KNOWN_BACKENDS:
            _LOGGER.warning(
                "%s: ignoring unknown recorded backend %r; treating as 'pip'.",
                self.venv_dir.name,
                recorded_backend,
            )
            recorded_backend = "pip"
        self.backend = recorded_backend
        self.read_metadata_version = input_dict.get("pipx_metadata_version")

    def _validate_before_write(self) -> None:
        if (
            self.main_package.package is None
            or self.main_package.package_or_url is None
            or not self.main_package.include_apps
        ):
            _LOGGER.debug(f"PipxMetadata corrupt:\n{self.to_dict()}")
            raise PipxError("Internal Error: PipxMetadata is corrupt, cannot write.")

    def write(self) -> None:
        self._validate_before_write()
        try:
            with open(self.venv_dir / PIPX_INFO_FILENAME, "w", encoding="utf-8") as pipx_metadata_fh:
                json.dump(
                    self.to_dict(),
                    pipx_metadata_fh,
                    indent=4,
                    sort_keys=True,
                    cls=JsonEncoderHandlesPath,
                )
        except OSError:
            _LOGGER.warning(
                pipx_wrap(
                    f"""
                    {hazard}  Unable to write {PIPX_INFO_FILENAME} to
                    {self.venv_dir}.  This may cause future pipx operations
                    involving {self.venv_dir.name} to fail or behave
                    incorrectly.
                    """,
                    subsequent_indent=" " * 4,
                )
            )

    def read(self, verbose: bool = False) -> None:
        try:
            with open(self.venv_dir / PIPX_INFO_FILENAME, "rb") as pipx_metadata_fh:
                payload: _RawMetadata = json.load(pipx_metadata_fh, object_hook=_json_decoder_object_hook)
                self.from_dict(payload)
        except OSError:  # Reset self if problem reading
            if verbose:
                _LOGGER.warning(
                    pipx_wrap(
                        f"""
                        {hazard}  Unable to read {PIPX_INFO_FILENAME} in
                        {self.venv_dir}.  This may cause this or future pipx
                        operations involving {self.venv_dir.name} to fail or
                        behave incorrectly.
                        """,
                        subsequent_indent=" " * 4,
                    )
                )
            return


def load_spec_file(path: Path) -> "_RawSpecFile":
    # Round-trips Path values through :class:`JsonEncoderHandlesPath`'s hook.
    with open(path, encoding="utf-8") as handle:
        payload: _RawSpecFile = json.load(handle, object_hook=_json_decoder_object_hook)
    return payload


__all__ = [
    "PIPX_INFO_FILENAME",
    "JsonEncoderHandlesPath",
    "PackageInfo",
    "PipxMetadata",
    "load_spec_file",
]



import logging
import os
import sys
from collections.abc import Sequence
from pathlib import Path
from typing import Final

from pipx import commands, paths
from pipx.colors import bold, red
from pipx.commands.common import expose_resources_globally
from pipx.constants import EXIT_CODE_OK, ExitCode
from pipx.emojis import sleep
from pipx.package_specifier import parse_specifier_for_upgrade
from pipx.shared_libs import shared_libs
from pipx.util import PipxError, pipx_wrap
from pipx.venv import Venv, VenvContainer

_LOGGER: Final[logging.Logger] = logging.getLogger(__name__)


def _upgrade_package(
    venv: Venv,
    package_name: str,
    pip_args: list[str],
    is_main_package: bool,
    force: bool,
    upgrading_all: bool,
) -> int:
    """Returns 1 if package version changed, 0 if same version"""
    package_metadata = venv.package_metadata[package_name]

    if package_metadata.package_or_url is None:
        raise PipxError(f"Internal Error: package {package_name} has corrupt pipx metadata.")
    elif package_metadata.pinned:
        if package_metadata.package != venv.main_package_name:
            _LOGGER.warning(
                f"Not upgrading pinned package {package_metadata.package} in venv {venv.name}. "
                f"Run `pipx unpin {venv.name}` to unpin it."
            )
        else:
            _LOGGER.warning(f"Not upgrading pinned package {venv.name}. Run `pipx unpin {venv.name}` to unpin it.")
        return 0

    package_or_url = parse_specifier_for_upgrade(package_metadata.package_or_url)
    old_version = package_metadata.package_version

    venv.upgrade_package(
        package_name,
        package_or_url,
        pip_args,
        include_dependencies=package_metadata.include_dependencies,
        include_apps=package_metadata.include_apps,
        is_main_package=is_main_package,
        suffix=package_metadata.suffix,
    )

    package_metadata = venv.package_metadata[package_name]

    display_name = f"{package_metadata.package}{package_metadata.suffix}"
    new_version = package_metadata.package_version

    if package_metadata.include_apps:
        expose_resources_globally(
            "app",
            paths.ctx.bin_dir,
            package_metadata.app_paths,
            force=force,
            suffix=package_metadata.suffix,
        )
        expose_resources_globally("man", paths.ctx.man_dir, package_metadata.man_paths, force=force)

    if package_metadata.include_dependencies:
        for app_paths in package_metadata.app_paths_of_dependencies.values():
            expose_resources_globally(
                "app",
                paths.ctx.bin_dir,
                app_paths,
                force=force,
                suffix=package_metadata.suffix,
            )
        for man_paths in package_metadata.man_paths_of_dependencies.values():
            expose_resources_globally("man", paths.ctx.man_dir, man_paths, force=force)

    if old_version == new_version:
        if upgrading_all:
            pass
        else:
            print(
                pipx_wrap(
                    f"""
                    {display_name} is already at latest version {old_version}
                    (location: {venv.root!s})
                    """
                )
            )
        return 0
    else:
        print(
            pipx_wrap(
                f"""
                upgraded package {display_name} from {old_version} to
                {new_version} (location: {venv.root!s})
                """
            )
        )
        return 1


def _upgrade_venv(
    venv_dir: Path,
    pip_args: list[str],
    verbose: bool,
    *,
    include_injected: bool,
    upgrading_all: bool,
    force: bool,
    install: bool = False,
    venv_args: list[str] | None = None,
    python: str | None = None,
    python_flag_passed: bool = False,
    backend: str | None = None,
    env_backend: str | None = None,
    venv: Venv | None = None,
    shared_libs_already_checked: bool = False,
) -> int:
    """Return number of packages whose versions changed.

    ``upgrade-all`` passes ``venv`` and ``shared_libs_already_checked=True``
    after its own pre-checks to avoid re-running them per venv.
    """
    if not venv_dir.is_dir():
        if install:
            if venv_args is None:
                venv_args = []
            commands.install(
                venv_dir=None,
                venv_args=venv_args,
                package_names=None,
                package_specs=[str(venv_dir).split(os.path.sep)[-1]],
                local_bin_dir=paths.ctx.bin_dir,
                local_man_dir=paths.ctx.man_dir,
                python=python,
                pip_args=pip_args,
                verbose=verbose,
                force=force,
                reinstall=False,
                include_dependencies=False,
                preinstall_packages=None,
                python_flag_passed=python_flag_passed,
                backend=backend,
                env_backend=env_backend,
            )
            return 0
        else:
            raise PipxError(
                f"""
                Package is not installed. Expected to find {venv_dir!s}, but it
                does not exist.
                """
            )

    if venv_args and not install:
        _LOGGER.info(f"Ignoring {', '.join(venv_args)} as not combined with --install")

    if python and not install:
        _LOGGER.info("Ignoring --python as not combined with --install")

    if venv is None:
        venv = Venv(venv_dir, verbose=verbose, backend=backend, env_backend=env_backend)
    if not pip_args:
        pip_args = venv.pipx_metadata.main_package.pip_args
    if not shared_libs_already_checked:
        venv.check_upgrade_shared_libs(pip_args=pip_args, verbose=verbose)

    if not venv.python_path.is_file():
        raise PipxError(
            f"Not upgrading {red(bold(venv_dir.name))}. It has an invalid python interpreter {venv.python_path}.\n"
            f"This usually happens after a system Python upgrade.\n"
            f"To fix, execute: pipx reinstall-all",
            wrap_message=False,
        )

    if not venv.package_metadata:
        raise PipxError(
            f"Not upgrading {red(bold(venv_dir.name))}. It has missing internal pipx metadata.\n"
            f"It was likely installed using a pipx version before 0.15.0.0.\n"
            f"Please uninstall and install this package to fix.",
            wrap_message=False,
        )

    # Upgrade shared libraries (pip, setuptools and wheel)
    venv.upgrade_packaging_libraries(pip_args)

    versions_updated = 0

    package_name = venv.main_package_name
    versions_updated += _upgrade_package(
        venv,
        package_name,
        pip_args,
        is_main_package=True,
        force=force,
        upgrading_all=upgrading_all,
    )

    if include_injected:
        for package_name in venv.package_metadata:
            if package_name == venv.main_package_name:
                continue
            injected_pip_args = pip_args or venv.package_metadata[package_name].pip_args
            versions_updated += _upgrade_package(
                venv,
                package_name,
                injected_pip_args,
                is_main_package=False,
                force=force,
                upgrading_all=upgrading_all,
            )

    return versions_updated


def upgrade(
    venv_dirs: dict[str, Path],
    python: str | None,
    pip_args: list[str],
    venv_args: list[str],
    verbose: bool,
    *,
    include_injected: bool,
    force: bool,
    install: bool,
    python_flag_passed: bool = False,
    backend: str | None = None,
    env_backend: str | None = None,
) -> ExitCode:
    """Return pipx exit code."""

    for venv_dir in venv_dirs.values():
        _ = _upgrade_venv(
            venv_dir,
            pip_args,
            verbose,
            include_injected=include_injected,
            upgrading_all=False,
            force=force,
            install=install,
            venv_args=venv_args,
            python=python,
            python_flag_passed=python_flag_passed,
            backend=backend,
            env_backend=env_backend,
        )

    # Any error in upgrade will raise PipxError (e.g. from venv.upgrade_package())
    return EXIT_CODE_OK


def upgrade_all(
    venv_container: VenvContainer,
    verbose: bool,
    *,
    pip_args: list[str],
    include_injected: bool,
    skip: Sequence[str],
    force: bool,
    python_flag_passed: bool = False,
    backend: str | None = None,
    env_backend: str | None = None,
) -> ExitCode:
    """Return pipx exit code."""
    failed: list[str] = []
    upgraded: list[str] = []

    for venv_dir in venv_container.iter_venv_dirs():
        # Cheap skip-list check first so we don't pay metadata read +
        # cross-backend warning + shared-libs health check on excluded venvs.
        if venv_dir.name in skip:
            continue
        venv = Venv(venv_dir, verbose=verbose, backend=backend, env_backend=env_backend)
        if "--editable" in venv.pipx_metadata.main_package.pip_args:
            continue
        venv.check_upgrade_shared_libs(pip_args=pip_args, verbose=verbose)
        try:
            versions_updated = _upgrade_venv(
                venv_dir,
                pip_args,
                verbose=verbose,
                include_injected=include_injected,
                upgrading_all=True,
                force=force,
                python_flag_passed=python_flag_passed,
                backend=backend,
                env_backend=env_backend,
                venv=venv,
                shared_libs_already_checked=True,
            )
            if versions_updated > 0:
                upgraded.append(venv_dir.name)
        except PipxError as e:
            print(e, file=sys.stderr)
            failed.append(venv_dir.name)
    if len(upgraded) == 0:
        print(f"No packages upgraded after running 'pipx upgrade-all' {sleep}")
    if len(failed) > 0:
        raise PipxError(f"The following package(s) failed to upgrade: {','.join(failed)}")
    # Any failure to install will raise PipxError, otherwise success
    return EXIT_CODE_OK


def upgrade_shared(
    verbose: bool,
    pip_args: list[str],
) -> ExitCode:
    # Always refreshes: the next pip-backed install needs a fresh shared-libs
    # venv even when all currently-installed venvs are uv-backed.
    shared_libs.upgrade(verbose=verbose, pip_args=pip_args, raises=True)
    return EXIT_CODE_OK


__all__ = [
    "upgrade",
    "upgrade_all",
    "upgrade_shared",
]



import logging
import os
import re
import sys
from collections.abc import Generator, Iterable
from pathlib import Path
from typing import Final

from packaging.utils import canonicalize_name

from pipx import paths
from pipx.backends import assert_not_pip_under_uv
from pipx.colors import bold
from pipx.commands.common import package_name_from_spec, run_post_install_actions
from pipx.constants import EXIT_CODE_INJECT_ERROR, EXIT_CODE_OK, ExitCode
from pipx.emojis import hazard, stars
from pipx.util import PipxError, pipx_wrap
from pipx.venv import Venv

_LOGGER: Final[logging.Logger] = logging.getLogger(__name__)

_COMMENT_RE: Final[re.Pattern[str]] = re.compile(r"(^|\s+)#.*$")


def inject_dep(
    venv_dir: Path,
    package_name: str | None,
    package_spec: str,
    pip_args: list[str],
    *,
    verbose: bool,
    include_apps: bool,
    include_dependencies: bool,
    force: bool,
    suffix: bool = False,
    backend: str | None = None,
    env_backend: str | None = None,
) -> bool:
    _LOGGER.debug("Injecting package %s", package_spec)

    if not venv_dir.exists() or next(venv_dir.iterdir(), None) is None:
        raise PipxError(
            f"""
            Can't inject {package_spec!r} into nonexistent Virtual Environment
            {venv_dir.name!r}. Be sure to install the package first with 'pipx
            install {venv_dir.name}' before injecting into it.
            """
        )

    venv = Venv(venv_dir, verbose=verbose, backend=backend, env_backend=env_backend)
    venv.check_upgrade_shared_libs(pip_args=pip_args, verbose=verbose)

    if not venv.package_metadata:
        raise PipxError(
            f"""
            Can't inject {package_spec!r} into Virtual Environment
            {venv.name!r}. {venv.name!r} has missing internal pipx metadata. It
            was likely installed using a pipx version before 0.15.0.0. Please
            uninstall and install {venv.name!r}, or reinstall-all to fix.
            """
        )

    # package_spec is anything pip-installable, including package_name, vcs spec,
    #   zip file, or tar.gz file.
    if package_name is None:
        package_name = package_name_from_spec(
            package_spec,
            os.fspath(venv.python_path),
            pip_args=pip_args,
            verbose=verbose,
            backend=venv.backend_name,
            env_backend=env_backend,
        )

    # Mirrors the install-side guard: dropping pip into a uv venv works for
    # ``pipx run`` but breaks anyone reaching for the venv's missing pip.
    assert_not_pip_under_uv(canonicalize_name(package_name), venv.backend_name)

    if not force and venv.has_package(package_name):
        _LOGGER.info("Package %s has already been injected", package_name)
        print(
            pipx_wrap(
                f"""
                {hazard} {package_name} already seems to be injected in {venv.name!r}.
                Not modifying existing installation in '{venv_dir}'.
                Pass '--force' to force installation.
                """
            )
        )
        return True

    if suffix:
        venv_suffix = venv.package_metadata[venv.main_package_name].suffix
    else:
        venv_suffix = ""
    venv.install_package(
        package_name=package_name,
        package_or_url=package_spec,
        pip_args=pip_args,
        install_only_pip_args=["--force-reinstall"] if force else None,
        include_dependencies=include_dependencies,
        include_apps=include_apps,
        is_main_package=False,
        suffix=venv_suffix,
    )
    if include_apps:
        run_post_install_actions(
            venv,
            package_name,
            paths.ctx.bin_dir,
            paths.ctx.man_dir,
            venv_dir,
            include_dependencies,
            force=force,
        )

    print(f"  injected package {bold(package_name)} into venv {bold(venv.name)}")
    print(f"done! {stars}", file=sys.stderr)

    # Any failure to install will raise PipxError, otherwise success
    return True


def inject(
    venv_dir: Path,
    package_specs: Iterable[str],
    requirement_files: Iterable[str],
    pip_args: list[str],
    *,
    verbose: bool,
    include_apps: bool,
    include_dependencies: bool,
    force: bool,
    suffix: bool = False,
    backend: str | None = None,
    env_backend: str | None = None,
) -> ExitCode:
    """Returns pipx exit code."""
    # Combined collection of package specifications
    packages = list(package_specs)
    for filename in requirement_files:
        packages.extend(parse_requirements(filename))

    # Remove duplicates and order deterministically
    packages = sorted(set(packages))

    if not packages:
        raise PipxError("No packages have been specified.")
    _LOGGER.info("Injecting packages: %r", packages)

    # Inject packages
    if not include_apps and include_dependencies:
        include_apps = True
    all_success = True
    for dependency in packages:
        all_success &= inject_dep(
            venv_dir,
            package_name=None,
            package_spec=dependency,
            pip_args=pip_args,
            verbose=verbose,
            include_apps=include_apps,
            include_dependencies=include_dependencies,
            force=force,
            suffix=suffix,
            backend=backend,
            env_backend=env_backend,
        )

    # Any failure to install will raise PipxError, otherwise success
    return EXIT_CODE_OK if all_success else EXIT_CODE_INJECT_ERROR


def parse_requirements(filename: str | os.PathLike) -> Generator[str, None, None]:
    """
    Extract package specifications from requirements file.

    Return all of the non-empty lines with comments removed.
    """
    # Based on https://github.com/pypa/pip/blob/main/src/pip/_internal/req/req_file.py
    with open(filename) as f:
        for line in f:
            # Strip comments and filter empty lines
            if pkgspec := _COMMENT_RE.sub("", line).strip():
                yield pkgspec


__all__ = [
    "inject",
    "inject_dep",
    "parse_requirements",
]



# PYTHON_ARGCOMPLETE_OK

"""The command line interface to pipx"""

import argparse
import logging
import logging.config
import os
import re
import shlex
import sys
import textwrap
import time
import urllib.parse
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Any, NoReturn

import argcomplete
import platformdirs
from packaging.utils import canonicalize_name

from pipx import commands, constants, paths
from pipx.animate import hide_cursor, show_cursor
from pipx.backends import KNOWN_BACKENDS, UV, env_default_backend, get_backend, resolve_backend_name
from pipx.colors import bold, green
from pipx.commands.environment import ENVIRONMENT_VALUE_CHOICES, ENVIRONMENT_VARIABLES
from pipx.constants import (
    _FETCH_MISSING_PYTHON_RAW,
    _FETCH_PYTHON,
    _FETCH_PYTHON_INVALID,
    _FETCH_PYTHON_RAW,
    EXIT_CODE_OK,
    EXIT_CODE_SPECIFIED_PYTHON_EXECUTABLE_NOT_FOUND,
    MINIMUM_PYTHON_VERSION,
    WINDOWS,
    ExitCode,
    FetchPythonOptions,
)
from pipx.emojis import hazard
from pipx.interpreter import (
    DEFAULT_PYTHON,
    InterpreterResolutionError,
    find_python_interpreter,
)
from pipx.package_specifier import valid_pypi_name
from pipx.util import PipxError, mkdir, pipx_wrap, rmdir
from pipx.venv import VenvContainer
from pipx.version import version as __version__

logger = logging.getLogger(__name__)

VenvCompleter = Callable[[str], list[str]]


def print_version() -> None:
    print(__version__)


def prog_name() -> str:
    try:
        prog = os.path.basename(sys.argv[0])
        if prog == "__main__.py":
            return f"{sys.executable} -m pipx"
        else:
            return prog
    except Exception:
        pass
    return "pipx"


SPEC_HELP = textwrap.dedent(
    """\
    The package name or specific installation source passed to pip.
    Runs `pip install -U SPEC`.
    For example `--spec mypackage==2.0.0` or `--spec  git+https://github.com/user/repo.git@branch`
    """
)

PIPX_DESCRIPTION = textwrap.dedent(
    f"""
    Install and execute apps from Python packages.

    Binaries can either be installed globally into isolated Virtual Environments
    or run directly in a temporary Virtual Environment.

    Virtual Environment location is {paths.ctx.venvs!s}.
    Symlinks to apps are placed in {paths.ctx.bin_dir!s}.
    Symlinks to manual pages are placed in {paths.ctx.man_dir!s}.

    """
)
PIPX_DESCRIPTION += pipx_wrap(
    """
    optional environment variables:
      PIPX_HOME              Overrides default pipx location. Virtual Environments will be installed to $PIPX_HOME/venvs.
      PIPX_GLOBAL_HOME       Used instead of PIPX_HOME when the `--global` option is given.
      PIPX_BIN_DIR           Overrides location of app installations. Apps are symlinked or copied here.
      PIPX_GLOBAL_BIN_DIR    Used instead of PIPX_BIN_DIR when the `--global` option is given.
      PIPX_MAN_DIR           Overrides location of manual pages installations. Manual pages are symlinked or copied here.
      PIPX_GLOBAL_MAN_DIR    Used instead of PIPX_MAN_DIR when the `--global` option is given.
      PIPX_DEFAULT_PYTHON    Overrides default python used for commands.
      PIPX_DEFAULT_BACKEND   Overrides which backend (`pip` or `uv`) is used for new venvs.
      PIPX_DISABLE_SHARED_LIBS_AUTO_UPGRADE
                            Skips automatic shared library upgrades.
      PIPX_USE_EMOJI         Overrides emoji behavior. Default value varies based on platform.
      PIPX_HOME_ALLOW_SPACE  Overrides default warning on spaces in the home path
    """,
    subsequent_indent=" " * 24,  # match the indent of argparse options
    keep_newlines=True,
)

DOC_DEFAULT_PYTHON = os.getenv("PIPX__DOC_DEFAULT_PYTHON", DEFAULT_PYTHON)

INSTALL_DESCRIPTION = textwrap.dedent(
    f"""
    The install command is the preferred way to globally install apps
    from python packages on your system. It creates an isolated virtual
    environment for the package, then ensures the package's apps are
    accessible on your $PATH. The package's manual pages installed in
    share/man/man[1-9] can be viewed with man on an operating system where
    it is available and the path in the environment variable `PIPX_MAN_DIR`
    (default: {paths.DEFAULT_PIPX_MAN_DIR}) is in the man search path
    ($MANPATH).

    The result: apps you can run from anywhere, located in packages
    you can cleanly upgrade or uninstall. Guaranteed to not have
    dependency version conflicts or interfere with your OS's python
    packages. 'sudo' is not required to do this.

    pipx install PACKAGE_SPEC ...
    pipx install --python PYTHON PACKAGE_SPEC
    pipx install VCS_URL
    pipx install ./LOCAL_PATH
    pipx install ZIP_FILE
    pipx install TAR_GZ_FILE

    The PACKAGE_SPEC argument is passed directly to `pip install`.

    Virtual Environments will be installed to `$PIPX_HOME/venvs`.
    The default pipx home location is {paths.DEFAULT_PIPX_HOME} and can
    be overridden by setting the environment variable `PIPX_HOME`.
    If the `--global` option is used, the default pipx home location
    instead is {paths.DEFAULT_PIPX_GLOBAL_HOME} and can be overridden
    by setting the environment variable `PIPX_GLOBAL_HOME`.

    The default app location is {paths.DEFAULT_PIPX_BIN_DIR} and can be
    overridden by setting the environment variable `PIPX_BIN_DIR`.
    If the `--global` option is used, the default app location instead
    is {paths.DEFAULT_PIPX_GLOBAL_BIN_DIR} and can be overridden by
    setting the environment variable `PIPX_GLOBAL_BIN_DIR`.

    The default manual pages location is {paths.DEFAULT_PIPX_MAN_DIR} and
    can be overridden by setting the environment variable `PIPX_MAN_DIR`.
    If the `--global` option is used, the default manual pages location
    instead is {paths.DEFAULT_PIPX_GLOBAL_MAN_DIR} and can be overridden
    by setting the environment variable `PIPX_GLOBAL_MAN_DIR`.

    The default python executable used to install a package is
    {DOC_DEFAULT_PYTHON} and can be overridden
    by setting the environment variable `PIPX_DEFAULT_PYTHON`.
    """
)


class LineWrapRawTextHelpFormatter(argparse.RawDescriptionHelpFormatter):
    def _split_lines(self, text: str, width: int) -> list[str]:
        text = self._whitespace_matcher.sub(" ", text).strip()
        return textwrap.wrap(text, width)


class InstalledVenvsCompleter:
    def __init__(self, venv_container: VenvContainer) -> None:
        self.packages = [str(p.name) for p in sorted(venv_container.iter_venv_dirs())]

    def use(self, prefix: str, **kwargs: Any) -> list[str]:
        return [f"{prefix}{x[len(prefix) :]}" for x in self.packages if x.startswith(canonicalize_name(prefix))]


def get_pip_args(parsed_args: dict[str, str]) -> list[str]:
    pip_args: list[str] = []
    if parsed_args.get("index_url"):
        pip_args += ["--index-url", parsed_args["index_url"]]

    if parsed_args.get("pip_args"):
        # Stripping the single quote that can be parsed from several shells
        pip_args_striped = parsed_args["pip_args"].strip("'")
        pip_args += shlex.split(pip_args_striped, posix=not WINDOWS)

    # make sure --editable is last because it needs to be right before
    #   package specification
    if parsed_args.get("editable"):
        pip_args += ["--editable"]
    return pip_args


def get_runpip_args(pip_args: list[str]) -> list[str]:
    if len(pip_args) != 1:
        return pip_args

    candidate = pip_args[0]
    # Allow a single quoted string like "install black".
    if not any(char.isspace() for char in candidate):
        return pip_args

    split_args = shlex.split(candidate, posix=not WINDOWS)
    return split_args or pip_args


def get_venv_args(parsed_args: dict[str, str]) -> list[str]:
    venv_args: list[str] = []
    if parsed_args.get("system_site_packages"):
        venv_args += ["--system-site-packages"]
    return venv_args


def get_backend_arg(parsed_args: dict[str, str]) -> str | None:
    return parsed_args.get("backend") or None


def package_is_url(package: str, raise_error: bool = True) -> bool:
    url_parse_package = urllib.parse.urlparse(package)
    if url_parse_package.scheme and url_parse_package.netloc:
        if not raise_error:
            return True
        raise PipxError("Package cannot be a URL. A valid package name should be passed instead.")
    return False


def package_is_path(package: str):
    if os.path.sep in package:
        raise PipxError(
            pipx_wrap(
                f"""
                Error: '{package}' looks like a path.
                Expected the name of an installed package.
                """
            )
        )


def _validate_fetch_python() -> None:
    if _FETCH_PYTHON_INVALID:
        valid = ", ".join(str(option) for option in FetchPythonOptions)
        raise PipxError(f"PIPX_FETCH_PYTHON must be unset or one of: {valid}.")
    if _FETCH_MISSING_PYTHON_RAW is not None:
        if _FETCH_PYTHON_RAW is not None:
            raise PipxError("Setting both PIPX_FETCH_MISSING_PYTHON and PIPX_FETCH_PYTHON is invalid.")
        print(
            f"{hazard} PIPX_FETCH_MISSING_PYTHON is deprecated; "
            f'use PIPX_FETCH_PYTHON="{FetchPythonOptions.MISSING}" instead.',
            file=sys.stderr,
        )


def run_pipx_command(args: argparse.Namespace) -> ExitCode:
    if "package" in args:
        package_is_url(args.package)
        package_is_path(args.package)
    if "packages" in args:
        for package in args.packages:
            package_is_url(package)
            package_is_path(package)

    spec: str | None = getattr(args, "spec", None)
    if "package" in args and spec is not None:
        if package_is_url(spec, raise_error=False) and "#egg=" not in spec:
            spec = f"{spec}#egg={args.package}"

    python = DEFAULT_PYTHON
    python_flag_passed = False
    if "python" in args:
        python_flag_passed = bool(args.python)
        try:
            python = find_python_interpreter(args.python or DEFAULT_PYTHON, fetch_python=args.fetch_python)
        except InterpreterResolutionError as e:
            logger.debug("Failed to resolve interpreter:", exc_info=True)
            print(pipx_wrap(f"{hazard} {e}", subsequent_indent=" " * 4))
            return EXIT_CODE_SPECIFIED_PYTHON_EXECUTABLE_NOT_FOUND

    cli_backend = get_backend_arg(vars(args))
    env_backend = env_default_backend()
    _validate_backend_available(cli_backend, env_backend)

    ctx = DispatchContext(
        verbose=bool(args.verbose) if "verbose" in args else False,
        pip_args=get_pip_args(vars(args)),
        venv_args=get_venv_args(vars(args)),
        venv_container=VenvContainer(paths.ctx.venvs),
        skip_list=[canonicalize_name(x) for x in args.skip] if "skip" in args else [],
        python=python,
        python_flag_passed=python_flag_passed,
        spec=spec,
        backend=cli_backend,
        env_backend=env_backend,
    )
    return args.func(args, ctx)


def _validate_backend_available(cli_backend: str | None, env_backend: str | None) -> None:
    """Eagerly surface a missing uv when ``--backend uv`` was passed.

    Env-source and auto-detect stay silent so diagnostic commands (which never
    instantiate a backend) keep working, and so unrelated PATH ``uv`` doesn't
    pay a ``uv --version`` subprocess on every command.
    """
    name, source = resolve_backend_name(cli_value=cli_backend, env_value=env_backend)
    logger.info(f"Backend resolved to {name} (source: {source})")
    if name == UV and source == "cli":
        get_backend(name)


@dataclass(frozen=True)
class DispatchContext:
    verbose: bool
    pip_args: list[str]
    venv_args: list[str]
    venv_container: VenvContainer
    skip_list: list[str]
    python: str
    python_flag_passed: bool
    spec: str | None
    backend: str | None = None
    env_backend: str | None = None

    @property
    def effective_backend(self) -> str | None:
        # Use ``backend`` and ``env_backend`` separately when building a
        # :class:`pipx.venv.Venv`; this only collapses them for code paths
        # without metadata to consult.
        return self.backend or self.env_backend


def add_pip_venv_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--system-site-packages",
        action="store_true",
        help="Give the virtual environment access to the system site-packages dir.",
    )
    parser.add_argument("--index-url", "-i", help="Base URL of Python Package Index")
    parser.add_argument(
        "--editable",
        "-e",
        help="Install a project in editable mode",
        action="store_true",
    )
    parser.add_argument(
        "--pip-args",
        help=(
            "Arbitrary pip arguments to pass directly to pip install/upgrade commands. "
            "Under the uv backend a small subset is translated (--index-url, --extra-index-url, "
            "--find-links, --pre); other flags raise an error so behaviour stays explicit."
        ),
    )


def add_backend_arg(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--backend",
        choices=list(KNOWN_BACKENDS),
        default=None,
        help=(
            "Backend used for creating venvs and managing packages. Defaults to "
            "PIPX_DEFAULT_BACKEND when set, else 'uv' when uv is available "
            "(via the `pipx[uv]` extra or on PATH), else 'pip'. Existing venvs "
            "always honor their recorded backend unless this flag is given."
        ),
    )


def add_include_dependencies(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("--include-deps", help="Include apps of dependent packages", action="store_true")


class _DeprecatedFetchMissingPython(argparse.Action):
    def __init__(self, option_strings: list[str], dest: str, **kwargs: Any) -> None:
        super().__init__(option_strings, dest, nargs=0, **kwargs)

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: str | None = None,
    ) -> None:
        if not {"--help", "-h"}.intersection(sys.argv):
            print(
                f"{hazard} {option_string} is deprecated; use --fetch-python={FetchPythonOptions.MISSING} instead.",
                file=sys.stderr,
            )
        setattr(namespace, self.dest, FetchPythonOptions.MISSING)


def add_python_options(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--python",
        help=(
            "Python to install with. Possible values can be the executable name (python3.11), "
            "the version of an available system Python or to pass to py launcher (3.11), "
            f"or the full path to the executable. Requires Python {MINIMUM_PYTHON_VERSION} or above."
        ),
    )
    fetch_python_group = parser.add_mutually_exclusive_group()
    fetch_python_group.add_argument(
        "--fetch-python",
        type=FetchPythonOptions,
        choices=list(FetchPythonOptions),
        default=_FETCH_PYTHON,
        help=(
            f"When to fetch a standalone Python build from python-build-standalone. "
            f"'{FetchPythonOptions.ALWAYS}' always downloads (ignores any system Python); "
            f"'{FetchPythonOptions.MISSING}' downloads only when the requested version is not found locally; "
            f"'{FetchPythonOptions.NEVER}' never downloads. "
            "Defaults to PIPX_FETCH_PYTHON, or 'never'."
        ),
    )
    fetch_python_group.add_argument(
        "--fetch-missing-python",
        dest="fetch_python",
        action=_DeprecatedFetchMissingPython,
        help="Deprecated: alias for --fetch-python=missing.",
    )


def _add_install(subparsers: argparse._SubParsersAction, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "install",
        help="Install a package",
        formatter_class=LineWrapRawTextHelpFormatter,
        description=INSTALL_DESCRIPTION,
        parents=[shared_parser],
    )
    p.add_argument("package_spec", help="package name(s) or pip installation spec(s)", nargs="+")
    add_include_dependencies(p)
    p.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Modify existing virtual environment and files in PIPX_BIN_DIR and PIPX_MAN_DIR",
    )
    p.add_argument(
        "--suffix",
        default="",
        help="Optional suffix for virtual environment and executable names.",
    )
    add_python_options(p)
    p.add_argument(
        "--preinstall",
        action="append",
        help=(
            "Optional package to be installed into the Virtual Environment before "
            "installing the main package. Use this flag multiple times if you want to preinstall multiple packages."
        ),
    )
    add_pip_venv_args(p)
    add_backend_arg(p)
    p.set_defaults(func=_cmd_install)


def _cmd_install(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.install(
        None,
        None,
        args.package_spec,
        paths.ctx.bin_dir,
        paths.ctx.man_dir,
        ctx.python,
        ctx.pip_args,
        ctx.venv_args,
        ctx.verbose,
        force=args.force,
        reinstall=False,
        include_dependencies=args.include_deps,
        preinstall_packages=args.preinstall,
        suffix=args.suffix,
        python_flag_passed=ctx.python_flag_passed,
        backend=ctx.backend,
        env_backend=ctx.env_backend,
    )


def _add_install_all(subparsers: argparse._SubParsersAction, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "install-all",
        help="Install all packages",
        formatter_class=LineWrapRawTextHelpFormatter,
        description="Installs all the packages according to spec metadata file.",
        parents=[shared_parser],
    )
    p.add_argument("spec_metadata_file", help="Spec metadata file generated from pipx list --json")
    p.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Modify existing virtual environment and files in PIPX_BIN_DIR and PIPX_MAN_DIR",
    )
    add_python_options(p)
    add_pip_venv_args(p)
    add_backend_arg(p)
    p.set_defaults(func=_cmd_install_all)


def _cmd_install_all(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.install_all(
        args.spec_metadata_file,
        paths.ctx.bin_dir,
        paths.ctx.man_dir,
        ctx.python,
        ctx.pip_args,
        ctx.venv_args,
        ctx.verbose,
        force=args.force,
        backend=ctx.backend,
        env_backend=ctx.env_backend,
    )


def _add_inject(subparsers, venv_completer: VenvCompleter, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "inject",
        help="Install packages into an existing Virtual Environment",
        description="Installs packages to an existing pipx-managed virtual environment.",
        parents=[shared_parser],
    )
    p.add_argument(
        "package",
        help="Name of the existing pipx-managed Virtual Environment to inject into",
    ).completer = venv_completer
    p.add_argument(
        "dependencies",
        nargs="*",
        help="the packages to inject into the Virtual Environment--either package name or pip package spec",
    )
    p.add_argument(
        "-r",
        "--requirement",
        dest="requirements",
        action="append",
        default=[],
        metavar="file",
        help=(
            "file containing the packages to inject into the Virtual Environment--"
            "one package name or pip package spec per line. "
            "May be specified multiple times."
        ),
    )
    p.add_argument(
        "--include-apps",
        action="store_true",
        help="Add apps from the injected packages onto your PATH and expose their manual pages",
    )
    p.add_argument(
        "--include-deps",
        help="Include apps of dependent packages. Implies --include-apps",
        action="store_true",
    )
    add_pip_venv_args(p)
    p.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Modify existing virtual environment and files in PIPX_BIN_DIR and PIPX_MAN_DIR",
    )
    p.add_argument(
        "--with-suffix",
        action="store_true",
        help="Add the suffix (if given) of the Virtual Environment to the packages to inject",
    )
    add_backend_arg(p)
    p.set_defaults(func=_cmd_inject)


def _cmd_inject(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.inject(
        _venv_dir(args, ctx),
        args.dependencies,
        args.requirements,
        ctx.pip_args,
        verbose=ctx.verbose,
        include_apps=args.include_apps,
        include_dependencies=args.include_deps,
        force=args.force,
        suffix=args.with_suffix,
        backend=ctx.backend,
        env_backend=ctx.env_backend,
    )


def _add_uninject(subparsers, venv_completer: VenvCompleter, shared_parser: argparse.ArgumentParser):
    p = subparsers.add_parser(
        "uninject",
        help="Uninstall injected packages from an existing Virtual Environment",
        description="Uninstalls injected packages from an existing pipx-managed virtual environment.",
        parents=[shared_parser],
    )
    p.add_argument(
        "package",
        help="Name of the existing pipx-managed Virtual Environment to inject into",
    ).completer = venv_completer
    p.add_argument(
        "dependencies",
        nargs="+",
        help="the package names to uninject from the Virtual Environment",
    )
    p.add_argument(
        "--leave-deps",
        action="store_true",
        help="Only uninstall the main injected package but leave its dependencies installed.",
    )
    p.set_defaults(func=_cmd_uninject)


def _cmd_uninject(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.uninject(
        _venv_dir(args, ctx),
        args.dependencies,
        local_bin_dir=paths.ctx.bin_dir,
        local_man_dir=paths.ctx.man_dir,
        leave_deps=args.leave_deps,
        verbose=ctx.verbose,
    )


def _add_pin(subparsers, venv_completer: VenvCompleter, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "pin",
        help="Pin the specified package to prevent it from being upgraded",
        description="Pin the specified package to prevent it from being upgraded",
        parents=[shared_parser],
    )
    p.add_argument("package", help="Installed package to pin")
    p.add_argument(
        "--injected-only",
        action="store_true",
        help=(
            "Pin injected packages in venv only, so that they will not be upgraded during upgrade operations. "
            "Note that this should not be passed if you wish to pin both main package and injected packages."
        ),
    )
    p.add_argument(
        "--skip",
        nargs="+",
        default=[],
        help="Skip these packages. Implies `--injected-only`.",
    )
    p.set_defaults(func=_cmd_pin)


def _cmd_pin(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.pin(_venv_dir(args, ctx), ctx.verbose, ctx.skip_list, args.injected_only)


def _add_unpin(subparsers, venv_completer: VenvCompleter, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "unpin",
        help="Unpin the specified package",
        description="Unpin the specified package and all injected packages in its venv to allow them to be upgraded",
        parents=[shared_parser],
    )
    p.add_argument("package", help="Installed package to unpin")
    p.set_defaults(func=_cmd_unpin)


def _cmd_unpin(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.unpin(_venv_dir(args, ctx), ctx.verbose)


def _add_upgrade(subparsers, venv_completer: VenvCompleter, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "upgrade",
        help="Upgrade a package",
        description="Upgrade package(s) in pipx-managed Virtual Environment(s) by running 'pip install --upgrade PACKAGE'",
        parents=[shared_parser],
    )
    p.add_argument("packages", help="package names(s) to upgrade", nargs="+").completer = venv_completer
    p.add_argument(
        "--include-injected",
        action="store_true",
        help="Also upgrade packages injected into the main app's environment",
    )
    p.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Modify existing virtual environment and files in PIPX_BIN_DIR and PIPX_MAN_DIR",
    )
    add_pip_venv_args(p)
    p.add_argument(
        "--install",
        action="store_true",
        help="Install package spec if missing",
    )
    add_python_options(p)
    add_backend_arg(p)
    p.set_defaults(func=_cmd_upgrade)


def _cmd_upgrade(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.upgrade(
        _venv_dirs(args, ctx),
        ctx.python,
        ctx.pip_args,
        ctx.venv_args,
        ctx.verbose,
        include_injected=args.include_injected,
        force=args.force,
        install=args.install,
        python_flag_passed=ctx.python_flag_passed,
        backend=ctx.backend,
        env_backend=ctx.env_backend,
    )


def _add_upgrade_all(subparsers: argparse._SubParsersAction, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "upgrade-all",
        help="Upgrade all packages. Runs `pip install -U <pkgname>` for each package.",
        description="Upgrades all packages within their virtual environments by running 'pip install --upgrade PACKAGE'",
        parents=[shared_parser],
    )
    p.add_argument(
        "--include-injected",
        action="store_true",
        help="Also upgrade packages injected into the main app's environment",
    )
    p.add_argument("--skip", nargs="+", default=[], help="skip these packages")
    p.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Modify existing virtual environment and files in PIPX_BIN_DIR and PIPX_MAN_DIR",
    )
    add_pip_venv_args(p)
    add_backend_arg(p)
    p.set_defaults(func=_cmd_upgrade_all)


def _cmd_upgrade_all(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.upgrade_all(
        ctx.venv_container,
        ctx.verbose,
        include_injected=args.include_injected,
        skip=ctx.skip_list,
        force=args.force,
        pip_args=ctx.pip_args,
        python_flag_passed=ctx.python_flag_passed,
        backend=ctx.backend,
        env_backend=ctx.env_backend,
    )


def _add_upgrade_shared(subparsers: argparse._SubParsersAction, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "upgrade-shared",
        help="Upgrade shared libraries.",
        description="Upgrade shared libraries.",
        parents=[shared_parser],
    )
    p.add_argument(
        "--pip-args",
        help="Arbitrary pip arguments to pass directly to pip install/upgrade commands",
    )
    p.set_defaults(func=_cmd_upgrade_shared)


def _cmd_upgrade_shared(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.upgrade_shared(ctx.verbose, ctx.pip_args)


def _add_uninstall(subparsers, venv_completer: VenvCompleter, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "uninstall",
        help="Uninstall a package",
        description="Uninstalls a pipx-managed Virtual Environment by deleting it and any files that point to its apps.",
        parents=[shared_parser],
    )
    p.add_argument("package").completer = venv_completer
    p.set_defaults(func=_cmd_uninstall)


def _cmd_uninstall(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.uninstall(_venv_dir(args, ctx), paths.ctx.bin_dir, paths.ctx.man_dir, ctx.verbose)


def _add_uninstall_all(subparsers: argparse._SubParsersAction, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "uninstall-all",
        help="Uninstall all packages",
        description="Uninstall all pipx-managed packages",
        parents=[shared_parser],
    )
    p.set_defaults(func=_cmd_uninstall_all)


def _cmd_uninstall_all(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.uninstall_all(ctx.venv_container, paths.ctx.bin_dir, paths.ctx.man_dir, ctx.verbose)


def _add_reinstall(subparsers, venv_completer: VenvCompleter, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "reinstall",
        formatter_class=LineWrapRawTextHelpFormatter,
        help="Reinstall a package",
        description=textwrap.dedent(
            """
            Reinstalls a package.

            Package is uninstalled, then installed with pipx install PACKAGE
            with the same options used in the original install of PACKAGE.

            """
        ),
        parents=[shared_parser],
    )
    p.add_argument("package").completer = venv_completer
    add_python_options(p)
    add_backend_arg(p)
    p.set_defaults(func=_cmd_reinstall)


def _cmd_reinstall(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.reinstall(
        venv_dir=_venv_dir(args, ctx),
        local_bin_dir=paths.ctx.bin_dir,
        local_man_dir=paths.ctx.man_dir,
        python=ctx.python,
        verbose=ctx.verbose,
        python_flag_passed=ctx.python_flag_passed,
        backend=ctx.backend,
        env_backend=ctx.env_backend,
    )


def _add_reinstall_all(subparsers: argparse._SubParsersAction, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "reinstall-all",
        formatter_class=LineWrapRawTextHelpFormatter,
        help="Reinstall all packages",
        description=textwrap.dedent(
            """
            Reinstalls all packages.

            Packages are uninstalled, then installed with pipx install PACKAGE
            with the same options used in the original install of PACKAGE.
            This is useful if you upgraded to a new version of Python and want
            all your packages to use the latest as well.

            """
        ),
        parents=[shared_parser],
    )
    add_python_options(p)
    p.add_argument("--skip", nargs="+", default=[], help="skip these packages")
    add_backend_arg(p)
    p.set_defaults(func=_cmd_reinstall_all)


def _cmd_reinstall_all(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.reinstall_all(
        ctx.venv_container,
        paths.ctx.bin_dir,
        paths.ctx.man_dir,
        ctx.python,
        ctx.verbose,
        skip=ctx.skip_list,
        python_flag_passed=ctx.python_flag_passed,
        backend=ctx.backend,
        env_backend=ctx.env_backend,
    )


def _add_list(subparsers: argparse._SubParsersAction, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "list",
        help="List installed packages",
        description="List packages and apps installed with pipx",
        parents=[shared_parser],
    )
    p.add_argument(
        "--include-injected",
        action="store_true",
        help="Show packages injected into the main app's environment",
    )
    g = p.add_mutually_exclusive_group()
    g.add_argument("--json", action="store_true", help="Output rich data in json format.")
    g.add_argument("--short", action="store_true", help="List packages only.")
    g.add_argument(
        "--pinned",
        action="store_true",
        help="List pinned packages only. Pass --include-injected at the same time to list injected packages that were pinned.",
    )
    g.add_argument("--skip-maintenance", action="store_true", help="(deprecated) No-op")
    p.set_defaults(func=_cmd_list)


def _cmd_list(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.list_packages(ctx.venv_container, args.include_injected, args.json, args.short, args.pinned)


def _add_interpreter(
    subparsers: argparse._SubParsersAction, shared_parser: argparse.ArgumentParser
) -> argparse.ArgumentParser:
    p: argparse.ArgumentParser = subparsers.add_parser(
        "interpreter",
        help="Interact with interpreters managed by pipx",
        description="Interact with interpreters managed by pipx",
        parents=[shared_parser],
    )
    s = p.add_subparsers(
        title="subcommands",
        description="Get help for commands with pipx interpreter COMMAND --help",
        dest="interpreter_command",
    )
    list_p = s.add_parser("list", help="List available interpreters", description="List available interpreters")
    prune_p = s.add_parser("prune", help="Prune unused interpreters", description="Prune unused interpreters")
    upgrade_p = s.add_parser(
        "upgrade",
        help="Upgrade installed interpreters to the latest available micro/patch version",
        description="Upgrade installed interpreters to the latest available micro/patch version",
    )
    list_p.set_defaults(func=_cmd_interpreter_list)
    prune_p.set_defaults(func=_cmd_interpreter_prune)
    upgrade_p.set_defaults(func=_cmd_interpreter_upgrade)
    p.set_defaults(func=_make_print_help(p))
    return p


def _cmd_interpreter_list(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.list_interpreters(ctx.venv_container)


def _cmd_interpreter_prune(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.prune_interpreters(ctx.venv_container)


def _cmd_interpreter_upgrade(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.upgrade_interpreters(ctx.venv_container, ctx.verbose)


def _add_run(subparsers: argparse._SubParsersAction, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "run",
        formatter_class=LineWrapRawTextHelpFormatter,
        help=(
            "Download the latest version of a package to a temporary virtual environment, "
            "then run an app from it. Also compatible with local `__pypackages__` "
            "directory (experimental)."
        ),
        description=textwrap.dedent(
            f"""
            Download the latest version of a package to a temporary virtual environment,
            then run an app from it. The environment will be cached
            and reused for up to {constants.TEMP_VENV_EXPIRATION_THRESHOLD_DAYS} days. This
            means subsequent calls to 'run' for the same package will be faster
            since they can reuse the cached Virtual Environment.

            In support of PEP 582 'run' will use apps found in a local __pypackages__
            directory, if present. Please note that this behavior is experimental,
            and acts as a companion tool to pythonloc. It may be modified or
            removed in the future. See https://github.com/cs01/pythonloc.
            """
        ),
        parents=[shared_parser],
    )
    p.add_argument(
        "--no-cache",
        action="store_true",
        help="Do not reuse cached virtual environment if it exists",
    )
    p.add_argument(
        "--no-path-check",
        action="store_true",
        help="Do not check whether the app is already on PATH",
    )
    p.add_argument(
        "app_with_args",
        metavar="app ...",
        nargs=argparse.REMAINDER,
        help="app/package name and any arguments to be passed to it",
        default=[],
    )
    p.add_argument("--path", action="store_true", help="Interpret app name as a local path")
    p.add_argument(
        "--pypackages",
        action="store_true",
        help="Require app to be run from local __pypackages__ directory",
    )
    p.add_argument(
        "--with",
        dest="with_",
        action="append",
        default=[],
        help="Extra dependencies to add to the temporary environment",
    )
    p.add_argument("--spec", help=SPEC_HELP)
    add_python_options(p)
    add_pip_venv_args(p)
    add_backend_arg(p)
    p.set_defaults(subparser=p, func=_cmd_run)

    # modify usage text to show required app argument
    p.usage = re.sub(r"^usage: ", "", p.format_usage())
    # add a double-dash to usage text to show requirement before app
    p.usage = re.sub(r"\.\.\.", "app ...", p.usage)


def _cmd_run(args: argparse.Namespace, ctx: DispatchContext) -> NoReturn:
    commands.run(
        args.app_with_args[0],
        ctx.spec,
        args.with_,
        args.path,
        args.app_with_args[1:],
        ctx.python,
        ctx.pip_args,
        ctx.venv_args,
        args.pypackages,
        ctx.verbose,
        not args.no_cache,
        no_path_check=args.no_path_check,
        backend=ctx.backend,
        env_backend=ctx.env_backend,
    )


def _add_runpip(subparsers, venv_completer: VenvCompleter, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "runpip",
        help="Run pip in an existing pipx-managed Virtual Environment",
        description="Run pip in an existing pipx-managed Virtual Environment",
        parents=[shared_parser],
    )
    p.add_argument(
        "package",
        help="Name of the existing pipx-managed Virtual Environment to run pip in",
    ).completer = venv_completer
    p.add_argument(
        "pipargs",
        nargs=argparse.REMAINDER,
        default=[],
        help="Arguments to forward to pip command",
    )
    p.set_defaults(func=_cmd_runpip)


def _cmd_runpip(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.run_pip(args.package, _venv_dir(args, ctx), get_runpip_args(args.pipargs), ctx.verbose)


def _add_ensurepath(subparsers: argparse._SubParsersAction, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "ensurepath",
        help=("Ensure directories necessary for pipx operation are in your PATH environment variable."),
        description=(
            "Ensure directory where pipx stores apps is in your "
            "PATH environment variable. Also if pipx was installed via "
            "`pip install --user`, ensure pipx itself is in your PATH. "
            "Note that running this may modify "
            "your shell's configuration file(s) such as '~/.bashrc'."
        ),
        parents=[shared_parser],
    )
    p.add_argument(
        "--prepend",
        action="store_true",
        help=(
            "Prepend directories to your PATH instead of appending. "
            "This is useful if you want to prioritize pipx apps over system apps."
        ),
    )
    p.add_argument(
        "--force",
        "-f",
        action="store_true",
        help=(
            "Add text to your shell's config file even if it looks like your "
            "PATH already contains paths to pipx and pipx-install apps."
        ),
    )
    p.add_argument(
        "--all-shells",
        action="store_true",
        help=("Add directories to PATH in all shells instead of just the current one."),
    )
    p.add_argument(
        "--dry-run",
        action="store_true",
        help=("Report what would be added to PATH without modifying PATH or any shell configuration file."),
    )
    p.set_defaults(func=_cmd_ensurepath)


def _cmd_ensurepath(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    try:
        return commands.ensure_pipx_paths(
            prepend=args.prepend, force=args.force, all_shells=args.all_shells, dry_run=args.dry_run
        )
    except Exception as e:
        logger.debug("Uncaught Exception:", exc_info=True)
        raise PipxError(str(e), wrap_message=False) from None


def _add_environment(subparsers: argparse._SubParsersAction, shared_parser: argparse.ArgumentParser) -> None:
    p = subparsers.add_parser(
        "environment",
        formatter_class=LineWrapRawTextHelpFormatter,
        help="Print a list of environment variables and paths used by pipx.",
        description=textwrap.dedent(
            """
            Prints the names and current values of environment variables used by pipx,
            followed by internal pipx variables which are derived from the environment
            variables and platform specific default values.

            Available variables:
            """
        )
        + textwrap.fill(", ".join(ENVIRONMENT_VARIABLES), break_long_words=False),
        parents=[shared_parser],
    )
    p.add_argument(
        "--value",
        "-V",
        choices=ENVIRONMENT_VALUE_CHOICES,
        metavar="VARIABLE",
        help="Print the value of the variable.",
    )
    p.set_defaults(func=_cmd_environment)


def _cmd_environment(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    return commands.environment(value=args.value)


def _venv_dir(args: argparse.Namespace, ctx: DispatchContext) -> Path:
    venv_dir = ctx.venv_container.get_venv_dir(valid_pypi_name(args.package) or args.package)
    logger.info(f"Virtual Environment location is {venv_dir}")
    return venv_dir


def _venv_dirs(args: argparse.Namespace, ctx: DispatchContext) -> dict[str, Path]:
    venv_dirs = {pkg: ctx.venv_container.get_venv_dir(valid_pypi_name(pkg) or pkg) for pkg in args.packages}
    venv_dirs_msg = "\n".join(f"- {key} : {value}" for key, value in venv_dirs.items())
    logger.info(f"Virtual Environment locations are:\n{venv_dirs_msg}")
    return venv_dirs


def _make_print_help(
    target_parser: argparse.ArgumentParser,
) -> Callable[[argparse.Namespace, DispatchContext], ExitCode]:
    def _print_help(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
        target_parser.print_help()
        return EXIT_CODE_OK

    return _print_help


def get_command_parser() -> tuple[argparse.ArgumentParser, dict[str, argparse.ArgumentParser]]:
    venv_container = VenvContainer(paths.ctx.venvs)

    completer_venvs = InstalledVenvsCompleter(venv_container)

    shared_parser = argparse.ArgumentParser(add_help=False)

    shared_parser.add_argument(
        "--quiet",
        "-q",
        action="count",
        default=0,
        help=(
            "Give less output. May be used multiple times corresponding to the"
            " ERROR and CRITICAL logging levels. The count maxes out at 2."
        ),
    )

    shared_parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help=(
            "Give more output. May be used multiple times corresponding to the"
            " INFO, DEBUG and NOTSET logging levels. The count maxes out at 3."
        ),
    )

    if not constants.WINDOWS:
        shared_parser.add_argument(
            "--global",
            action="store_true",
            dest="is_global",
            help="Perform action globally for all users.",
        )

    parser = argparse.ArgumentParser(
        prog=prog_name(),
        formatter_class=LineWrapRawTextHelpFormatter,
        description=PIPX_DESCRIPTION,
    )
    parser.man_short_description = PIPX_DESCRIPTION.splitlines()[1]  # type: ignore[attr-defined]

    subparsers = parser.add_subparsers(dest="command", description="Get help for commands with pipx COMMAND --help")

    subparsers_with_subcommands = {}
    _add_install(subparsers, shared_parser)
    _add_install_all(subparsers, shared_parser)
    _add_uninject(subparsers, completer_venvs.use, shared_parser)
    _add_inject(subparsers, completer_venvs.use, shared_parser)
    _add_pin(subparsers, completer_venvs.use, shared_parser)
    _add_unpin(subparsers, completer_venvs.use, shared_parser)
    _add_upgrade(subparsers, completer_venvs.use, shared_parser)
    _add_upgrade_all(subparsers, shared_parser)
    _add_upgrade_shared(subparsers, shared_parser)
    _add_uninstall(subparsers, completer_venvs.use, shared_parser)
    _add_uninstall_all(subparsers, shared_parser)
    _add_reinstall(subparsers, completer_venvs.use, shared_parser)
    _add_reinstall_all(subparsers, shared_parser)
    _add_list(subparsers, shared_parser)
    subparsers_with_subcommands["interpreter"] = _add_interpreter(subparsers, shared_parser)
    _add_run(subparsers, shared_parser)
    _add_runpip(subparsers, completer_venvs.use, shared_parser)
    _add_ensurepath(subparsers, shared_parser)
    _add_environment(subparsers, shared_parser)

    parser.add_argument("--version", action="store_true", help="Print version and exit")
    completions_p = subparsers.add_parser(
        "completions",
        help="Print instructions on enabling shell completions for pipx",
        description="Print instructions on enabling shell completions for pipx",
        parents=[shared_parser],
    )
    completions_p.set_defaults(func=_cmd_completions)
    help_p = subparsers.add_parser(
        "help",
        help="Show help for pipx or a command",
        description="Show help for pipx or a command",
        parents=[shared_parser],
    )
    help_p.set_defaults(func=_make_print_help(parser))
    return parser, subparsers_with_subcommands


def _cmd_completions(args: argparse.Namespace, ctx: DispatchContext) -> ExitCode:
    print(constants.completion_instructions)
    return ExitCode(0)


def delete_oldest_logs(file_list: list[Path], keep_number: int) -> None:
    file_list = sorted(file_list)
    if len(file_list) > keep_number:
        for existing_file in file_list[:-keep_number]:
            try:
                existing_file.unlink()
            except FileNotFoundError:
                pass


def _setup_log_file(pipx_log_dir: Path | None = None) -> Path:
    max_logs = int(os.getenv("PIPX_MAX_LOGS", 10))
    pipx_log_dir = pipx_log_dir or paths.ctx.logs
    # don't use utils.mkdir, to prevent emission of log message
    pipx_log_dir.mkdir(parents=True, exist_ok=True)

    delete_oldest_logs(list(pipx_log_dir.glob("cmd_*[0-9].log")), max_logs)
    delete_oldest_logs(list(pipx_log_dir.glob("cmd_*_pip_errors.log")), max_logs)

    datetime_str = time.strftime("%Y-%m-%d_%H.%M.%S")
    log_file = pipx_log_dir / f"cmd_{datetime_str}.log"
    counter = 1
    while log_file.exists() and counter < 10:
        log_file = pipx_log_dir / f"cmd_{datetime_str}_{counter}.log"
        counter += 1

    log_file.touch()

    return log_file


def setup_log_file() -> Path:
    try:
        return _setup_log_file()
    except PermissionError:
        return _setup_log_file(platformdirs.user_log_path("pipx"))


def setup_logging(verbose: int) -> None:
    pipx_str = (sys.stdout and sys.stdout.isatty() and bold(green("pipx >"))) or "pipx >"
    paths.ctx.log_file = setup_log_file()

    # Determine logging level, a value between 0 and 50
    level_number = min(max(0, logging.WARNING - 10 * verbose), 50)

    level = logging.getLevelName(level_number)

    # "incremental" is False so previous pytest tests don't accumulate handlers
    logging_config = {
        "version": 1,
        "formatters": {
            "stream_nonverbose": {
                "class": "logging.Formatter",
                "format": "{message}",
                "style": "{",
            },
            "stream_verbose": {
                "class": "logging.Formatter",
                "format": pipx_str + "({funcName}:{lineno}): {message}",
                "style": "{",
            },
            "file": {
                "class": "logging.Formatter",
                "format": "{relativeCreated: >8.1f}ms ({funcName}:{lineno}): {message}",
                "style": "{",
            },
        },
        "handlers": {
            "stream": {
                "class": "logging.StreamHandler",
                "formatter": "stream_verbose" if verbose else "stream_nonverbose",
                "level": level,
            },
            "file": {
                "class": "logging.FileHandler",
                "formatter": "file",
                "filename": str(paths.ctx.log_file),
                "encoding": "utf-8",
                "level": "DEBUG",
            },
        },
        "loggers": {"pipx": {"handlers": ["stream", "file"], "level": "DEBUG"}},
        "incremental": False,
    }
    logging.config.dictConfig(logging_config)


def setup(args: argparse.Namespace) -> None:
    if "version" in args and args.version:
        print_version()
        sys.exit(0)

    if not constants.WINDOWS and getattr(args, "is_global", False):
        paths.ctx.make_global()

    verbose = getattr(args, "verbose", 0) - getattr(args, "quiet", 0)

    setup_logging(verbose)

    logger.debug(f"{time.strftime('%Y-%m-%d %H:%M:%S')}")
    logger.debug(f"{' '.join(sys.argv)}")
    logger.info(f"pipx version is {__version__}")
    logger.info(f"Default python interpreter is '{DEFAULT_PYTHON}'")

    mkdir(paths.ctx.venvs)
    mkdir(paths.ctx.bin_dir)
    mkdir(paths.ctx.man_dir)
    mkdir(paths.ctx.venv_cache)
    mkdir(paths.ctx.standalone_python_cachedir)

    for cachedir in [
        paths.ctx.venv_cache,
        paths.ctx.standalone_python_cachedir,
    ]:
        cachedir_tag = cachedir / "CACHEDIR.TAG"
        if not cachedir_tag.exists():
            logger.debug("Adding CACHEDIR.TAG to cache directory")
            signature = (
                "Signature: 8a477f597d28d172789f06886806bc55\n"
                "# This file is a cache directory tag created by pipx.\n"
                "# For information about cache directory tags, see:\n"
                "#       https://bford.info/cachedir/\n"
            )
            with open(cachedir_tag, "w") as file:
                file.write(signature)

    rmdir(paths.ctx.trash, False)

    old_pipx_venv_location = paths.ctx.venvs / "pipx-app"
    if old_pipx_venv_location.exists():
        logger.warning(
            pipx_wrap(
                f"""
                {hazard}  A virtual environment for pipx was detected at
                {old_pipx_venv_location!s}. The 'pipx-app' package has been
                renamed back to 'pipx'
                (https://github.com/pypa/pipx/issues/82).
                """,
                subsequent_indent=" " * 4,
            )
        )


def check_args(parsed_pipx_args: argparse.Namespace) -> None:
    if parsed_pipx_args.command == "run":
        # we manually discard a first -- because using nargs=argparse.REMAINDER
        #   will not do it automatically
        if parsed_pipx_args.app_with_args and parsed_pipx_args.app_with_args[0] == "--":
            parsed_pipx_args.app_with_args.pop(0)
        # since we would like app to be required but not in a separate argparse
        #   add_argument, we implement our own missing required arg error
        if not parsed_pipx_args.app_with_args:
            parsed_pipx_args.subparser.error("the following arguments are required: app")


def normalize_help_command(args: list[str]) -> list[str]:
    if args and args[0] == "help":
        if len(args) == 1:
            return ["--help"]
        return args[1:] + ["--help"]
    return args


def _get_subparser(parser: argparse.ArgumentParser, command: str) -> argparse.ArgumentParser:
    subparsers_action = next(action for action in parser._actions if isinstance(action, argparse._SubParsersAction))
    return subparsers_action.choices[command]


def parse_pipx_args(parser: argparse.ArgumentParser, args: list[str]) -> argparse.Namespace:
    args = normalize_help_command(args)
    if args and args[0] == "inject":
        parsed_args = _get_subparser(parser, "inject").parse_intermixed_args(args[1:])
        parsed_args.command = "inject"
        return parsed_args
    return parser.parse_args(args)


def cli() -> ExitCode:
    """Entry point from command line"""
    try:
        hide_cursor()
        parser, _ = get_command_parser()
        argcomplete.autocomplete(parser, always_complete_options=False)
        parsed_pipx_args = parse_pipx_args(parser, sys.argv[1:])
        _validate_fetch_python()
        setup(parsed_pipx_args)
        check_args(parsed_pipx_args)
        if not parsed_pipx_args.command:
            parser.print_help()
            return ExitCode(1)
        return run_pipx_command(parsed_pipx_args)
    except PipxError as e:
        print(str(e), file=sys.stderr)
        logger.debug(f"PipxError: {e}", exc_info=True)
        return ExitCode(1)
    except KeyboardInterrupt:
        return ExitCode(1)
    except Exception:
        logger.debug("Uncaught Exception:", exc_info=True)
        raise
    finally:
        logger.debug("pipx finished.")
        show_cursor()


if __name__ == "__main__":
    sys.exit(cli())
